<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\Entity\ConfigEntityListBuilder;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a listing of AccessPolicy type entities.
 */
class AccessPolicyListBuilder extends ConfigEntityListBuilder {

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $information;

  /**
   * The access rule plugin manager.
   *
   * @var \Drupal\access_policy\AccessRulePluginManager
   */
  protected $accessRuleManager;

  /**
   * Constructs a new AccessPolicyListBuilder.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param \Drupal\Core\Entity\EntityStorageInterface $storage
   *   The entity storage class.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\AccessPolicyInformation $information
   *   Access policy information service.
   * @param \Drupal\access_policy\AccessRulePluginManager $access_rule_manager
   *   The access rule plugin manaager.
   */
  public function __construct(EntityTypeInterface $entity_type, EntityStorageInterface $storage, MessengerInterface $messenger, EntityTypeManagerInterface $entity_type_manager, AccessPolicyInformation $information, AccessRulePluginManager $access_rule_manager) {
    parent::__construct($entity_type, $storage);
    $this->messenger = $messenger;
    $this->entityTypeManager = $entity_type_manager;
    $this->accessRuleManager = $access_rule_manager;
    $this->accessPolicyInfo = $information;
    $this->entityTypes = $this->accessPolicyInfo->getAllEnabledEntityTypes();
  }

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity_type.manager')->getStorage($entity_type->id()),
      $container->get('messenger'),
      $container->get('entity_type.manager'),
      $container->get('access_policy.information'),
      $container->get('plugin.manager.access_rule'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'access_policy_admin_overview';
  }

  /**
   * {@inheritdoc}
   */
  public function render() {
    $build = [];
    foreach ($this->load() as $entity_type => $entities) {
      if (!isset($this->entityTypes[$entity_type])) {
        continue;
      }

      $build[$entity_type] = $this->renderTable($this->entityTypes[$entity_type], $entities);
    }
    if (empty($build)) {
      $build = $this->renderEmptyTable();
    }
    return $build;
  }

  /**
   * Render an empty table when no results found.
   *
   * @return array
   *   Empty table render array.
   */
  public function renderEmptyTable() {
    $table = [
      '#type' => 'table',
      '#header' => $this->buildHeader(),
      '#rows' => [],
      '#empty' => $this->t('No Access policies yet'),
    ];
    return $table;

  }

  /**
   * Render the access policy table.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The current entity type.
   * @param array $entities
   *   The array of access policies.
   *
   * @return array
   *   The fully formed table render array.
   */
  public function renderTable(EntityTypeInterface $entity_type, array $entities) {
    $table = [
      '#prefix' => '<div id = "policies-' . $entity_type->id() . '"><h2>' . $entity_type->getLabel() . '</h2>',
      '#type' => 'table',
      '#header' => $this->buildHeader(),
      '#rows' => [],
      '#suffix' => '</div>',
    ];
    foreach ($entities as $entity) {
      if ($row = $this->buildRow($entity)) {
        $table['#rows'][$entity->id()] = $row;
      }
    }

    // Move content at the top.
    if ($entity_type->id() == 'node') {
      $table['#weight'] = -10;
    }

    $table['#rows']['_sort'][] = [
      'data' => [
        '#type' => 'link',
        '#url' => Url::fromRoute('access_policy.access_policy_sort_form', ['target_entity_type_id' => $entity_type->id()]),
        '#title' => $this->t('Rearrange'),
        '#attributes' => [
          'class' => ['use-ajax'],
          'data-dialog-type' => 'modal',
          'data-dialog-options' => Json::encode([
            'width' => 700,
          ]),
        ],
      ],
      'colspan' => count($table['#header']),
    ];

    return $table;
  }

  /**
   * {@inheritdoc}
   */
  public function buildHeader() {
    $header['label'] = $this->t('Name');
    $header['type'] = $this->t('Type');
    $header['target_bundles'] = $this->t('Bundles');
    $header['access_rules'] = $this->t('Access Rules');
    return $header + parent::buildHeader();
  }

  /**
   * {@inheritdoc}
   */
  public function buildRow(EntityInterface $entity) {
    $row['label'] = $entity->label();
    $row['type'] = $this->getPolicyTypeLabel($entity);
    $row['target_bundles'] = $this->getBundleLabels($entity, 5);
    $row['access_rules'] = $this->getAccessRuleLabels($entity, 2);
    return $row + parent::buildRow($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function load() {
    $entities = [];
    foreach (parent::load() as $entity) {
      $entities[$entity->getTargetEntityTypeId()][] = $entity;
    }
    return $entities;
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultOperations(EntityInterface $entity) {
    /** @var \Drupal\field\FieldConfigInterface $entity */
    $operations = parent::getDefaultOperations($entity);

    if (isset($operations['edit'])) {
      $operations['edit']['weight'] = 30;
      $operations['edit']['title'] = $this->t('Configure');
    }

    return $operations;
  }

  /**
   * Get a list of entity type labels.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   *
   * @return string
   *   The list of enabled entity types.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getPolicyTypeLabel(EntityInterface $entity) {
    $type = $entity->getTypePlugin();
    $definition = $type->getPluginDefinition();
    return $definition['label'];
  }

  /**
   * Get a list of bundle labels.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The current access policy.
   * @param int $limit
   *   Set limit of how many bundle labels to show.
   *
   * @return string
   *   The list of enabled bundles.
   */
  public function getBundleLabels(AccessPolicyInterface $access_policy, $limit = NULL) {
    $entity_type = $access_policy->getTargetEntityTypeId();
    $bundleInfo = \Drupal::service('entity_type.bundle.info');

    $labels = [];
    $bundle_info = $bundleInfo->getBundleInfo($entity_type);
    foreach ($bundle_info as $id => $bundle) {
      if ($this->isSelectedBundle($access_policy, $id)) {
        $labels[] = $bundle['label'];
      }
    }

    return $this->renderLabels($labels, $limit);
  }

  /**
   * Determine whether a bundle is enabled for this access policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The current access policy.
   * @param string $bundle
   *   The name of the bundle.
   *
   * @return bool
   *   TRUE if it is enabled; FALSE otherwise.
   */
  public function isSelectedBundle(AccessPolicyInterface $access_policy, $bundle) {
    $bundles = $access_policy->getTargetBundles();
    if (in_array($bundle, $bundles)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Get a list of entity type labels.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param int $limit
   *   Set limit of how many labels to show.
   *
   * @return string
   *   The list of enabled entity types.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getAccessRuleLabels(EntityInterface $entity, $limit = NULL) {
    $rules = $this->accessRuleManager->getAccessRulePlugins($entity);
    $labels = [];
    if (!empty($rules)) {
      foreach ($rules as $plugin) {
        $labels[] = $plugin->getDefinition()->getLabel();
      }
    }

    return $this->renderLabels($labels, $limit);
  }

  /**
   * Render the labels.
   *
   * @param array $labels
   *   Array of labels.
   * @param int $limit
   *   Set limit of how many labels to show.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The rendered labels.
   */
  public function renderLabels(array $labels, $limit = NULL) {
    if (empty($labels)) {
      return $this->t('None');
    }

    $truncated = FALSE;
    $total = count($labels);
    if ($limit && count($labels) > $limit) {
      $labels = array_slice($labels, 0, $limit);
      $truncated = TRUE;
    }

    $string = implode(', ', $labels);
    if ($truncated) {
      $diff_count = $total - $limit;
      $string .= "… (" . $diff_count . " more)";
    }

    return $string;
  }

}
