<?php

namespace Drupal\access_policy;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\AlterableInterface;
use Drupal\Core\Database\Query\ConditionInterface;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AccountProxy;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class for handling Access Policy query alter.
 *
 * This generates a dynamic query based on how access policies and access
 * rules are configured. The logic of the select query is always going to follow
 * this format.
 *
 * SELECT column FROM node_field_data
 * WHERE
 * access_policy = 'policy' AND (field = 0 AND/OR field2 = 3...) OR
 * access_policy = 'policy 2' AND (field = 1 AND/OR field2 = 3...)
 *
 * @package Drupal\access_policy
 */
class AccessPolicyQueryAlter implements ContainerInjectionInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The access rule plugin manager.
   *
   * @var \Drupal\access_policy\AccessRulePluginManager
   */
  protected $accessRuleManager;

  /**
   * The current_user service.
   *
   * @var \Drupal\Core\Session\AccountProxy
   */
  protected $currentUser;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The entity type.
   *
   * @var \Drupal\Core\Entity\EntityTypeInterface
   */
  protected $entityType;

  /**
   * The query object.
   *
   * @var \Drupal\Core\Database\Query\AlterableInterface
   */
  protected $query;

  /**
   * The data table name.
   *
   * @var string
   */
  protected $dataTable;

  /**
   * The base table name.
   *
   * @var string
   */
  protected $baseTable;

  /**
   * The base field name.
   *
   * @var string
   */
  protected $baseField;

  /**
   * Array of access rules grouped by access policy.
   *
   * @var array
   */
  protected $accessRules = [];

  /**
   * AccessPolicyQueryAlter constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\AccessRulePluginManager $access_rule_manager
   *   The access rule plugin manager.
   * @param \Drupal\Core\Session\AccountProxy $current_user
   *   The current_user service.
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AccessRulePluginManager $access_rule_manager, AccountProxy $current_user, Connection $connection) {
    $this->entityTypeManager = $entity_type_manager;
    $this->accessRuleManager = $access_rule_manager;
    $this->currentUser = $this->entityTypeManager->getStorage('user')->load($current_user->id());
    $this->connection = $connection;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.access_rule'),
      $container->get('current_user'),
      $container->get('database')
    );
  }

  /**
   * Alter a query.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *
   * @see hook_query_alter()
   */
  public function queryAlter(AlterableInterface $query, EntityTypeInterface $entity_type) {
    $this->query = $query;
    $this->entityType = $entity_type;

    $altered = $query->getMetaData('access_policy');
    if (!$altered) {
      $this->baseTable = $this->getBaseTable($this->query, $entity_type);

      // If the base table does not match the entity type then do not proceed
      // with query. This can happen in the case of books or forums which both
      // tag the query with node_access.
      if ($this->isEntityBaseTable($this->baseTable, $this->query, $entity_type)) {

        // The table is valid, we can safely get the base field.
        $this->baseField = $entity_type->getKey('id');

        $this->ensureTable($this->query, $entity_type);

        // Only query the access policies of the current active entity type.
        $access_policies = $this->entityTypeManager->getStorage('access_policy')
          ->loadByProperties(['target_entity_type_id' => $entity_type->id()]);

        $or_group = $this->query->orConditionGroup();
        foreach ($access_policies as $access_policy) {
          // If the user can't view it then ensure the data table so that we
          // can compare against the status column.
          if (!$this->userCanViewUnpublished($access_policy, $this->currentUser) && empty($this->dataTable)) {
            $this->dataTable = $this->ensureDataTable($this->query, $entity_type);
          }

          // @todo this currently doesn't fully support AccessPolicyType.
          // For example, it can't accommodate "rules only" policies.
          /** @var \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy */
          if ($this->userCanView($access_policy, $this->currentUser)) {
            $this->accessPolicyQueryAlter($or_group, $access_policy);
          }
        }

        // If any policies exist then we need to add the OR group for
        // unassigned and empty policies. If no policies exist then we
        // shouldn't modify the query at all.
        if (!empty($access_policies)) {
          $column = $this->getAccessPolicyColumn();
          $or_group->condition($column, NULL, 'IS NULL');
          $or_group->condition($column, '');
          $this->query->condition($or_group);
        }

        // Flag this query as altered by access policy.
        $this->query->addMetaData('access_policy', TRUE);
      }
    }
  }

  /**
   * Get the current base table.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @return string
   *   The current base table.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getBaseTable(AlterableInterface $query, EntityTypeInterface $entity_type) {
    $tables = $query->getTables();
    $base_table = $query->getMetaData('base_table');
    if (!$base_table) {
      /** @var \Drupal\Core\Entity\Sql\DefaultTableMapping $table_mapping */
      $table_mapping = $this->entityTypeManager->getStorage($entity_type->id())->getTableMapping();
      $entity_base_tables = $table_mapping->getTableNames();
      $table_mapping->getDataTable();

      foreach ($tables as $table_info) {
        if (!($table_info instanceof SelectInterface)) {
          $table = $table_info['table'];
          // If the table is either the base table or data table then use
          // one of those.
          if ($table == $entity_type->getBaseTable() || $table == $entity_type->getDataTable()) {
            // If the table has an alias then make sure and use that instead.
            $base_table = $table_info['alias'] ?? $table;
            break;
          }
          // If one of the node base tables are in the query, add it to the list
          // of possible base tables to join against.
          if (in_array($table, $entity_base_tables) && is_null($table_info['join type'])) {
            $base_table = $table_info['alias'] ?? $table;
            break;
          }
        }
      }
    }

    return $base_table;
  }

  /**
   * Ensure the data table.
   *
   * This is only joined if the user does not have access to unpublished content
   * and if it's not joined already.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @return string
   *   The data table.
   */
  public function ensureDataTable(AlterableInterface $query, EntityTypeInterface $entity_type) {
    $tables = $query->getTables();

    $join = TRUE;
    $data_table = $entity_type->getDataTable();
    foreach ($tables as $table_info) {
      $table = $table_info['table'];
      // If it's already joined then do not join it again.
      if ($table == $entity_type->getDataTable()) {
        $join = FALSE;
        $data_table = $table_info['alias'] ?? $table;
      }
    }

    if ($join) {
      $query->innerJoin($data_table, $data_table, $data_table . '.' . $this->baseField . ' = ' . $this->baseTable . '.' . $this->baseField);
    }

    return $data_table;
  }

  /**
   * Determine whether the base table belongs to the entity.
   *
   * @param string $base_table
   *   The current base table.
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @return bool
   *   TRUE if the base table belongs to the entity; FALSE otherwise.
   */
  private function isEntityBaseTable($base_table, AlterableInterface $query, EntityTypeInterface $entity_type) {
    // The base table can sometimes be an alias. Grab the actual table name.
    $tables = $query->getTables();
    if (isset($tables[$base_table])) {
      $base_table = $tables[$base_table]['table'];

      /** @var \Drupal\Core\Entity\Sql\DefaultTableMapping $table_mapping */
      $table_mapping = $this->entityTypeManager->getStorage($entity_type->id())->getTableMapping();
      $entity_base_tables = $table_mapping->getTableNames();
      return in_array($base_table, $entity_base_tables);
    }

    return FALSE;
  }

  /**
   * Make sure that the access policy table has been joined properly.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   */
  public function ensureTable(AlterableInterface $query, EntityTypeInterface $entity_type) {
    $tables = $query->getTables();
    $join = TRUE;
    foreach ($tables as $table) {
      // Ensure that we don't join the table twice. this can happen when
      // filtering on views.
      if ($table['table'] == 'content_access_policy') {
        $join = FALSE;
        break;
      }
    }

    if ($join) {
      // Always join the access policy table.
      $join_condition = "content_access_policy.content_entity_id = " . $this->getBaseTablePlaceholder() . " AND content_access_policy.content_entity_type_id = '" . $entity_type->id() . "'";
      $query->leftJoin('content_access_policy', 'content_access_policy', $join_condition);
    }
  }

  /**
   * Alter a query based on Access Policies.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $orGroup
   *   The condition group.
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The Access Policy.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function accessPolicyQueryAlter(ConditionInterface $orGroup, AccessPolicyInterface $access_policy) {
    // The access policy will always have the logic of
    // OR access_policy = 'policy' AND (access rules AND/OR)
    $has_access_rules = $this->hasAccessRules($access_policy);
    if ($has_access_rules) {
      $operator = $access_policy->getAccessRuleOperator();
      switch ($operator) {
        case 'AND':
          $this->createAndGroupCondition($orGroup, $access_policy);
          break;

        case 'OR':
          $this->createOrGroupCondition($orGroup, $access_policy);
          break;
      }
    }

    // If no rules are supported or no query was altered then do a normal
    // condition without any AND group.
    if (!$has_access_rules || !$this->queryIsAltered($orGroup)) {
      $this->createConditionWithoutAccessRules($orGroup, $access_policy);
    }
  }

  /**
   * Get the base table placeholder.
   *
   * @return string
   *   The base table.
   */
  private function getBaseTablePlaceholder() {
    return $this->baseTable . '.' . $this->baseField;
  }

  /**
   * Determine whether the access policy has rules that should be queried.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   *
   * @return bool
   *   TRUE if it has access rules; FALSE otherwise.
   */
  private function hasAccessRules(AccessPolicyInterface $access_policy) {
    $is_query_enabled = $access_policy->isQueryEnabled();
    $has_rules = $access_policy->getTypePlugin()->supportsRules() && $access_policy->getAccessRules();
    $does_not_bypass = !$this->currentUser->hasPermission("bypass {$access_policy->id()} access rules");
    if ($is_query_enabled && $has_rules && $does_not_bypass) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Create the And group condition for access rule plugins.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $orGroup
   *   The condition group.
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The Access Policy.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function createAndGroupCondition(ConditionInterface $orGroup, AccessPolicyInterface $access_policy) {
    $access_rules = $this->getAccessRulePlugins($access_policy, ['query_alter' => TRUE]);
    $andGroup = $this->query->andConditionGroup();

    $this->accessRulesQueryAlter($andGroup, $access_rules);

    // If the query was actually altered then create the AND group and add the
    // conditions. This helps to prevent plugins without queries from breaking
    // the query.
    if ($this->queryIsAltered($andGroup)) {
      $andGroup->condition($this->getAccessPolicyColumn(), $access_policy->id());

      // If the user can't view unpublished content then require status=1.
      if (!$this->userCanViewUnpublished($access_policy, $this->currentUser)) {
        $status = $this->entityType->getKey('published');
        $andGroup->condition($this->dataTable . '.' . $status, TRUE);
      }

      $orGroup->condition($andGroup);
    }
  }

  /**
   * Create the OR group condition for access rule plugins.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $orGroup
   *   The condition group.
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The Access Policy.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function createOrGroupCondition(ConditionInterface $orGroup, AccessPolicyInterface $access_policy) {
    $access_rules = $this->getAccessRulePlugins($access_policy, ['query_alter' => TRUE]);
    $condGroup = $this->query->orConditionGroup();

    $this->accessRulesQueryAlter($condGroup, $access_rules);

    // If the query was actually altered then create the AND group and add the
    // OR conditions. This helps to prevent plugins without queries from
    // breaking the query.
    if ($this->queryIsAltered($condGroup)) {
      $andGroup = $this->query->andConditionGroup();
      $andGroup->condition($condGroup);
      $andGroup->condition($this->getAccessPolicyColumn(), $access_policy->id());

      // If the user can't view unpublished content then require status=1.
      if (!$this->userCanViewUnpublished($access_policy, $this->currentUser)) {
        $status = $this->entityType->getKey('published');
        $andGroup->condition($this->dataTable . '.' . $status, TRUE);
      }

      $orGroup->condition($andGroup);
    }
  }

  /**
   * Create condition for policy without access rules.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $orGroup
   *   The current or group.
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The current access policy.
   */
  private function createConditionWithoutAccessRules(ConditionInterface $orGroup, AccessPolicyInterface $access_policy) {
    // If the user can't view unpublished content then require status=1.
    if (!$this->userCanViewUnpublished($access_policy, $this->currentUser)) {
      $andGroup = $this->query->andConditionGroup();
      $andGroup->condition($this->getAccessPolicyColumn(), $access_policy->id());
      $status = $this->entityType->getKey('published');
      $andGroup->condition($this->dataTable . '.' . $status, TRUE);

      $orGroup->condition($andGroup);
    }
    else {
      $orGroup->condition($this->getAccessPolicyColumn(), $access_policy->id());
    }
  }

  /**
   * Determine whether the query was actually altered.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $condGroup
   *   The condition group.
   *
   * @return bool
   *   TRUE if the query was altered, FALSE otherwise.
   */
  private function queryIsAltered(ConditionInterface $condGroup) {
    foreach ($condGroup->conditions() as $key => $condition) {
      if (is_int($key)) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Alter a query based on Access Rules.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $condGroup
   *   The condition group.
   * @param array $access_rules
   *   The Access Rules.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function accessRulesQueryAlter(ConditionInterface $condGroup, array $access_rules) {
    foreach ($access_rules as $access_rule) {

      // If a handler is provided then use that, otherwise fallback to the
      // query alter directly on the access rule.
      $handler = $access_rule->getHandler('query_alter');
      if ($handler) {
        $select = new SqlQuery();
        $select->init($this->baseTable, $this->baseField);
        $handler->setQuery($select);
        $handler->query();
        $select = $handler->getQuery();
      }
      // @todo remove this once AccessRule::queryAlter has been removed.
      else {
        $select = new SelectQueryStub($this->connection, $this->baseTable);
        $select->addMetaData('base_table', $this->baseTable);
        $select->addMetaData('base_field', $this->baseField);
        $access_rule->queryAlter($select, $this->currentUser);
      }

      if ($select->isDistinct()) {
        $this->query->distinct();
      }

      $this->joinTablesFromQueryStub($this->query, $select);
      $this->addConditionsFromQueryStub($condGroup, $select);
    }
  }

  /**
   * Join tables from the query stub to the main query.
   *
   * This also ensures that any tables that have already been joined don't get
   * joined again as it could create duplicate results.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The main query.
   * @param \Drupal\access_policy\QueryInterface $queryStub
   *   The access policy query object.
   */
  public function joinTablesFromQueryStub(AlterableInterface $query, QueryInterface $queryStub) {
    $existing_tables = $query->getTables();
    $tables = $queryStub->getTables();
    $join_tables = array_filter($tables, function ($new_table) use ($existing_tables) {
      if (!is_null($new_table['join type'])) {
        $join_table = TRUE;
        // Make sure we don't join tables that have already been joined.
        foreach ($existing_tables as $table) {
          if ($new_table['table'] == $table['table'] && $new_table['alias'] == $table['alias']) {
            $join_table = FALSE;
            break;
          }
        }
        return $join_table;
      }
      return FALSE;
    });

    foreach ($join_tables as $table) {
      $query->addJoin($table['join type'], $table['table'], $table['alias'], $table['condition'], $table['arguments']);
    }
  }

  /**
   * Add conditions from a pseudo query.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $condGroup
   *   The condition group.
   * @param \Drupal\access_policy\QueryInterface $queryStub
   *   The access policy query object.
   */
  public function addConditionsFromQueryStub(ConditionInterface $condGroup, QueryInterface $queryStub) {
    $conditions = $queryStub->conditions();

    foreach ($conditions as $key => $condition) {
      if (is_int($key)) {
        $condGroup->condition($condition['field'], $condition['value'], $condition['operator']);
      }
    }
  }

  /**
   * Get Access Strategy Column Name based on the table name.
   *
   * @return string
   *   The column name.
   */
  protected function getAccessPolicyColumn(): string {
    return 'content_access_policy.access_policy';
  }

  /**
   * Determine whether the current user can view content with an access policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user has access; FALSE otherwise.
   */
  private function userCanView(AccessPolicyInterface $access_policy, AccountInterface $account) {
    // If the query is enabled for the access policy then check to see if
    // they have access first. If it's not enabled then they always have
    // access.
    if ($access_policy->isQueryEnabled() && $access_policy->getTypePlugin()->shouldValidatePermission('view')) {
      $entity_type = $access_policy->getTargetEntityType();
      $label = strtolower($entity_type->getLabel()->getUntranslatedString());
      $has_permission = $account->hasPermission("view {$access_policy->id()} " . $label);

      if (!$has_permission) {
        return FALSE;
      }
    }

    if ($this->hasAccessRules($access_policy)) {
      return $this->validateNonQueryAccessRules($access_policy, $account);
    }

    return TRUE;
  }

  /**
   * Validate the non query alter access rules.
   *
   * Access rules can validate against user/environmental attributes and entity
   * attributes. Only entity attributes actually need to alter the query. The
   * rest can go through normal access rule validation.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the access rule passes; FALSE otherwise.
   */
  private function validateNonQueryAccessRules(AccessPolicyInterface $access_policy, AccountInterface $account) {
    // Validate the non query alter access rules.
    $plugins = $this->getAccessRulePlugins($access_policy, ['validate' => TRUE]);
    if (!empty($plugins)) {
      $accessPolicyValidator = \Drupal::service('access_policy.validator');
      // @todo replace with an ExecutionContext object.
      $entity = $this->createStubEntity($access_policy);
      $is_valid = $accessPolicyValidator->validateAccessRulePlugins($plugins, $entity, $account, $access_policy->getAccessRuleOperator());
      if (!$is_valid) {
        // Only exit if this is an AND operator or all the access rules have
        // been validated.
        $has_query_alter_plugins = $this->getAccessRulePlugins($access_policy, ['query_alter' => TRUE]);
        if ($access_policy->getAccessRuleOperator() == 'AND' || !$has_query_alter_plugins) {
          return FALSE;
        }
      }
    }

    return TRUE;
  }

  /**
   * Get access rule plugins from a policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   * @param array $filter
   *   Filter by query alter or validate.
   *
   * @return array
   *   Array of access rules.
   */
  private function getAccessRulePlugins(AccessPolicyInterface $access_policy, array $filter = []) {
    if (!isset($this->accessRules[$access_policy->id()])) {
      $this->accessRules[$access_policy->id()] = [];

      $plugins = $this->accessRuleManager->getAccessRulePlugins($access_policy);
      foreach ($plugins as $id => $plugin) {
        $definition = $plugin->getDefinition();
        $settings = $definition->getSettings();

        // If query is enabled for this access rule then continue to fetch and
        // group them.
        if (isset($settings['query']) && $settings['query'] === TRUE) {
          $this->accessRules[$access_policy->id()][$id]['plugin'] = $plugin;

          // If the plugin group matches the current access policy entity type
          // then we can perform a query alter.
          if ($definition->getGroup() == $access_policy->getTargetEntityTypeId()) {
            $plugin_definition = $plugin->getPluginDefinition();
            // @todo remove this once AccessRule::queryAlter has been removed.
            $has_deprecated_method = method_exists($plugin, 'queryAlter');
            if (isset($plugin_definition['handlers']['query_alter']) || $has_deprecated_method) {
              $this->accessRules[$access_policy->id()][$id]['query_alter'] = TRUE;
            }
          }
          // If the group doesn't match the current access policy entity type
          // then we are not going to perform a query alter.
          elseif ($definition->getGroup() != $access_policy->getTargetEntityTypeId()) {
            $this->accessRules[$access_policy->id()][$id]['validate'] = TRUE;
          }
        }
      }
    }

    $access_rules = [];
    foreach ($this->accessRules[$access_policy->id()] as $item) {
      if (!empty($filter)) {
        $return = TRUE;
        foreach ($filter as $key => $value) {
          if (!isset($item[$key]) || (isset($item[$key]) && $item[$key] != $value)) {
            $return = FALSE;
          }
        }
        if ($return) {
          $access_rules[] = $item['plugin'];
        }
      }
      else {
        $access_rules[] = $item['plugin'];
      }
    }

    return $access_rules;
  }

  /**
   * Create a stub entity for validating non-query alter access rule.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The stub entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function createStubEntity(AccessPolicyInterface $access_policy) {
    // @todo replace this with an ExecutionContext object.
    $entity_type_id = $access_policy->getTargetEntityTypeId();
    // Grab one of the bundles. It doesn't really matter which one.
    $bundle = $access_policy->getTargetBundles()[0];

    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    $bundle_key = $entity_type->getKey('bundle');
    return $this->entityTypeManager->getStorage($entity_type_id)->create([
      $bundle_key => $bundle,
    ]);
  }

  /**
   * Determine whether the current user can view unpublished content.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user has access; FALSE otherwise.
   */
  private function userCanViewUnpublished(AccessPolicyInterface $access_policy, AccountInterface $account) {
    if ($access_policy->isQueryEnabled() && $access_policy->getTypePlugin()->shouldValidatePermission('view unpublished')) {
      $entity_type = $access_policy->getTargetEntityType();
      $label = strtolower($entity_type->getLabel()->getUntranslatedString());
      return $account->hasPermission("view any {$access_policy->id()} unpublished " . $label);
    }

    return TRUE;
  }

}
