<?php

namespace Drupal\access_policy;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Provides a listing of AccessPolicy type entities.
 */
class AccessPolicySelection implements AccessPolicySelectionInterface {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * The access policy validator.
   *
   * @var \Drupal\access_policy\AccessPolicyValidator
   */
  protected $accessPolicyValidator;

  /**
   * Constructs a new AccessPolicySelection service.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\AccessPolicyInformation $access_policy_information
   *   The access policy information service.
   * @param \Drupal\access_policy\AccessPolicyValidator $access_policy_validator
   *   The access policy validator.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AccessPolicyInformation $access_policy_information, AccessPolicyValidator $access_policy_validator) {
    $this->entityTypeManager = $entity_type_manager;
    $this->accessPolicyInformation = $access_policy_information;
    $this->accessPolicyValidator = $access_policy_validator;
  }

  /**
   * {@inheritdoc}
   */
  public function getAllowedAccessPolicies(EntityInterface $entity, AccountInterface $account) {
    $policies = $this->getAccessPoliciesSorted($entity->getEntityType(), $entity->bundle());
    $allowed = [];
    foreach ($policies as $policy) {
      if ($this->isAllowed($policy, $entity, $account)) {
        $allowed[] = $policy;
      }
    }

    return $allowed;
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultPolicy(EntityInterface $entity, AccountInterface $account) {
    $default_value = $this->getConfig()->get('default_policy');
    if ($default_value == 'first_available') {
      $policies = $this->getAccessPoliciesSorted($entity->getEntityType(), $entity->bundle());

      if (!empty($policies)) {
        foreach ($policies as $policy) {
          if ($this->isAllowed($policy, $entity, $account)) {
            return $policy;
          }
        }
      }
    }

    return FALSE;
  }

  /**
   * Determine whether a user is allowed to assign an access policy.
   *
   * @param \Drupal\Core\Entity\EntityInterface $policy
   *   The access policy entity.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user can use the policy; FALSE otherwise.
   */
  public function isAllowed(EntityInterface $policy, EntityInterface $entity, AccountInterface $account) {
    return $this->accessPolicyValidator->validateAccessPolicy($policy, $entity, $account, 'assign access policy');
  }

  /**
   * Get access policies sorted by ascending weight.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   * @param string $bundle
   *   The bundle.
   *
   * @return array
   *   Array of sorted access policies.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getAccessPoliciesSorted(EntityTypeInterface $entity_type, $bundle) {
    $policies = $this->accessPolicyInformation->getEnabledForEntitiesOfEntityType($entity_type, $bundle);
    $sorted = [];
    if (!empty($policies)) {
      foreach ($policies as $policy) {
        $sorted[$policy->id()] = $policy->getWeight();
      }
      asort($sorted);

      return $this->entityTypeManager->getStorage('access_policy')
        ->loadMultiple(array_keys($sorted));
    }

    return [];
  }

  /**
   * Determine whether empty access policies are allowed.
   *
   * @return bool
   *   TRUE if empty policies are allowed; FALSE otherwise.
   */
  public function emptyPoliciesAllowed() {
    if ($this->getConfig()->get('allow_empty')) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Get the config settings.
   *
   * Note: This is currently a workaround to get tests to work. Currently,
   * config settings don't get passed in through dependency injection.
   *
   * @return \Drupal\Core\Config\ImmutableConfig
   *   A configuration object.
   */
  private function getConfig() {
    return \Drupal::service('config.factory')->get('access_policy.settings');
  }

}
