<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;

/**
 * Service for validating access policies on entities.
 */
class AccessPolicyValidator implements AccessPolicyValidatorInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The content access policy manager service.
   *
   * @var \Drupal\access_policy\ContentAccessPolicyManager
   */
  protected $contentAccessPolicyManager;

  /**
   * The access rule plugin manager.
   *
   * @var \Drupal\access_policy\AccessRulePluginManager
   */
  protected $accessRulePluginManager;

  /**
   * Array of violations.
   *
   * @var array
   */
  protected $violations = [];

  /**
   * Constructs an AccessPolicyValidator object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\ContentAccessPolicyManager $content_access_policy_manager
   *   The content access policy manager.
   * @param \Drupal\access_policy\AccessRulePluginManager $access_rule_manager
   *   The access rule plugin manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, ContentAccessPolicyManager $content_access_policy_manager, AccessRulePluginManager $access_rule_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->contentAccessPolicyManager = $content_access_policy_manager;
    $this->accessRulePluginManager = $access_rule_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function validate(EntityInterface $entity, AccountInterface $account, string $operation) {
    // Make sure the violations are reset.
    $access_policy = $this->contentAccessPolicyManager->getAccessPolicy($entity);
    if (!empty($access_policy)) {
      return $this->validateAccessPolicy($access_policy, $entity, $account, $operation);
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function validateAccessPolicy(AccessPolicyInterface $access_policy, EntityInterface $entity, AccountInterface $account, $operation) {
    $this->violations = [];
    $type = $access_policy->getTypePlugin();

    if ($type->shouldValidatePermission($operation)) {
      $permission = $this->getAccessPolicyPermission($access_policy, $entity->getEntityType(), $operation);
      if (!$account->hasPermission($permission)) {
        $this->addViolation(sprintf('Missing permission: %s', $permission));
        return FALSE;
      }
    }

    if ($type->shouldValidateRules($operation)) {
      if (!$account->hasPermission('bypass ' . $access_policy->id() . ' access rules')) {
        $access_rules_check = $this->validateAccessRules($access_policy, $entity, $account, $operation);
        if (!$access_rules_check) {
          return FALSE;
        }
      }
    }

    return TRUE;
  }

  /**
   * Get the properly formed access policy permission.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy entity.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   * @param string $op
   *   The operation.
   *
   * @return string
   *   The permission string.
   */
  private function getAccessPolicyPermission(AccessPolicyInterface $access_policy, EntityTypeInterface $entity_type, $op) {
    $entity_type_label = strtolower($entity_type->getLabel()->getUntranslatedString());
    switch ($op) {
      case 'update':
        return 'edit ' . $access_policy->id() . ' ' . $entity_type_label;

      case 'assign access policy':
        return 'assign ' . $access_policy->id() . ' access policy';

      case 'view all revisions':
        return 'view all ' . $access_policy->id() . ' ' . $entity_type_label . ' revisions';

      case 'view unpublished':
        return 'view any ' . $access_policy->id() . ' unpublished ' . $entity_type_label;

      default:
        return $op . ' ' . $access_policy->id() . ' ' . $entity_type_label;
    }
  }

  /**
   * Validate Access Rules in a policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $policy
   *   The access policy.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The content entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   * @param string $operation
   *   The current operation.
   *
   * @return bool
   *   Return TRUE if valid, FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function validateAccessRules(AccessPolicyInterface $policy, EntityInterface $entity, AccountInterface $account, $operation) {
    $access_rules = $this->accessRulePluginManager->getApplicableAccessRulePlugins($policy, $entity);

    $type = $policy->getTypePlugin();
    $default_status = $type->ruleValidationEnabled($operation);
    $access_rules = $this->filterDisabledOperations($access_rules, $operation, $default_status);

    return $this->validateAccessRulePlugins($access_rules, $entity, $account, $policy->getAccessRuleOperator());
  }

  /**
   * Validate Access Rules plugins.
   *
   * @param array $access_rules
   *   Array of access rule plugins.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The content entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   * @param string $operator
   *   The current operator; default is AND.
   *
   * @return bool
   *   Return TRUE if valid, FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function validateAccessRulePlugins(array $access_rules, EntityInterface $entity, AccountInterface $account, $operator = 'AND') {
    $user = $this->entityTypeManager->getStorage('user')->load($account->id());
    $access = TRUE;

    foreach ($access_rules as $plugin) {
      // Compute the final value that will be validated against.
      // @todo add some other kind of initialization.
      $plugin->getValue();
      $access = $plugin->validate($entity, $user);

      switch ($operator) {
        // If all rules are required then fail immediately if any return false.
        case 'AND':
          if (!$access) {
            $this->addViolation(sprintf('Access rule failed: %s', $plugin->getDefinition()->getLabel()));
            return FALSE;
          }
          break;

        // If any rules are required then pass immediately if any return true.
        case 'OR':
          if ($access) {
            return TRUE;
          }
          break;
      }

      // Note that access rules with OR can still trigger a violation though
      // the final access is still granted.
      if (!$access) {
        $this->addViolation(sprintf('Access rule failed: %s', $plugin->getDefinition()->getLabel()));
      }

    }

    return $access;
  }

  /**
   * Filter out any rules that are not configured to execute for this operation.
   *
   * If the access rule does not have any operation overrides then it will
   * fall back to using the policy type operation settings.
   *
   * @param array $rules
   *   Array of access rules.
   * @param string $operation
   *   The current operation.
   * @param bool $default
   *   The default operation setting from the policy type.
   *
   * @return array
   *   Array of filtered access rule plugins.
   */
  private function filterDisabledOperations(array $rules, $operation, $default) {
    return array_filter($rules, function ($rule) use ($operation, $default) {
      $settings = $rule->getSettings();
      // If operations is not set then fallback to the default operation setting
      // from the policy type.
      if (!isset($settings['operations'])) {
        return $default;
      }

      if (in_array($operation, $settings['operations'])) {
        return TRUE;
      }
      return FALSE;
    });
  }

  /**
   * Add a violation message.
   *
   * @param string $message
   *   The violation message.
   */
  protected function addViolation($message) {
    return $this->violations[] = $message;
  }

  /**
   * Return any violation messages.
   *
   * @return array
   *   Array of violation messages.
   */
  public function getViolations() {
    return $this->violations;
  }

  /**
   * Determine whether there are any violations.
   *
   * @return bool
   *   TRUE if there are violations; FALSE otherwise.
   */
  public function hasViolations() {
    return !empty($this->violations);
  }

}
