<?php

namespace Drupal\access_policy;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Class to manage and lazy load cached access rule data.
 */
class AccessRuleData {

  use StringTranslationTrait;

  /**
   * The base cache ID to use.
   *
   * @var string
   */
  protected $baseCid = 'access_rule_data';

  /**
   * The cache backend to use.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cacheBackend;

  /**
   * The access rule data storage.
   *
   * @var array
   */
  protected $storage = [];

  /**
   * Stores a module manager to invoke hooks.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Constructs this AccessRuleData object.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   The cache backend to use.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler class to use for invoking hooks.
   */
  public function __construct(CacheBackendInterface $cache_backend, ModuleHandlerInterface $module_handler) {
    $this->cacheBackend = $cache_backend;
    $this->moduleHandler = $module_handler;
  }

  /**
   * Gets all access rule data.
   *
   * @return array
   *   An array of access rule data.
   */
  public function getAll() {
    if (empty($this->storage)) {
      $this->storage = $this->getData();
    }

    return $this->storage;
  }

  /**
   * Get all the access rules in a particular group.
   *
   * @param string $group
   *   The group.
   *
   * @return array
   *   The array of access rule data.
   */
  public function getGroup($group) {
    if (empty($this->storage)) {
      $this->storage = $this->getData();
    }
    $storage = [];
    foreach ($this->storage as $id => $item) {
      if (substr_count($id, $group . ':') > 0) {
        $storage[$id] = $item;
      }
    }

    return $storage;
  }

  /**
   * Gets data for a particular access rule.
   *
   * @param string $group
   *   The group.
   * @param string $key
   *   The key of the cache entry to retrieve.
   *
   * @return array
   *   An array of access rule data.
   *
   * @throws \Exception
   *   It will throw an exception if no data is found.
   */
  public function getRule($group, $key) {
    if (!isset($this->storage[$group . ':' . $key])) {
      $this->storage = $this->getData();
    }
    if ($this->dataExists($group, $key)) {
      return $this->storage[$group . ':' . $key];
    }

    return [];
  }

  /**
   * Determine whether the data exists.
   *
   * @param string $group
   *   The group.
   * @param string $key
   *   The key of the cache entry.
   *
   * @return bool
   *   TRUE if the data exists; an exception if it does not.
   */
  private function dataExists($group, $key) {
    if (empty($this->storage[$group . ':' . $key])) {
      $all_keys = array_keys($this->storage);
      $available = implode(', ', $all_keys);
      throw new \Exception(sprintf('Access rule data not found for "%1$s" in "%2$s". Available definitions include: %3$s.', $key, $group, $available));
    }
    return TRUE;
  }

  /**
   * Gets all data invoked by hook_views_data().
   *
   * This is requested from the cache before being rebuilt.
   *
   * @return array
   *   An array of all data.
   */
  protected function getData() {
    if ($data = $this->cacheGet($this->baseCid)) {
      return $data->data;
    }

    $data = [];
    $this->moduleHandler->invokeAllWith('access_rule_data', function (callable $hook, string $module) use (&$data) {
      $access_rule_data = $hook();

      // Make sure that each access rule has the provider.
      foreach ($access_rule_data as $type => $access_rules) {
        foreach ($access_rules as $id => $rule) {
          if (!isset($rule['provider'])) {
            $access_rule_data[$type][$id]['provider'] = $module;
          }
        }
      }

      if (is_array($access_rule_data)) {
        $data = NestedArray::mergeDeep($data, $access_rule_data);
      }
    });

    // Allow other modules to manipulate the data.
    $this->moduleHandler->alter('access_rule_data', $data);

    // Normalize the data into a flattened array.
    $storage = [];
    foreach ($data as $group => $items) {
      foreach ($items as $id => $item) {
        $storage[$group . ':' . $id] = $item;
      }
    }

    $this->cacheSet($this->baseCid, $storage);

    $this->storage = $storage;

    return $this->storage;
  }

  /**
   * Gets data from the cache backend.
   *
   * @param string $cid
   *   The cache ID to return.
   *
   * @return mixed
   *   The cached data, if any.
   */
  protected function cacheGet($cid) {
    return $this->cacheBackend->get($cid);
  }

  /**
   * Sets data to the cache backend.
   *
   * @param string $cid
   *   The cache ID to set.
   * @param mixed $data
   *   The data that will be cached.
   */
  protected function cacheSet($cid, $data) {
    return $this->cacheBackend->set($cid, $data, Cache::PERMANENT, ['access_rule_data', 'config:core.extension']);
  }

  /**
   * Clears the class storage and cache.
   */
  public function clear() {
    $this->storage = [];
    $this->cacheBackend->delete($this->baseCid);
  }

}
