<?php

namespace Drupal\access_policy\AccessRuleQueryHandler;

/**
 * Field value access rule query handler.
 */
class EntityFieldBase extends AccessRuleQueryHandlerBase {

  /**
   * {@inheritdoc}
   */
  public function query() {
    $this->ensureMyTable();

    $values = $this->getValue();

    // If empty_behavior is set to deny then require the values to be set.
    if ($this->definition->getSetting('empty_behavior') == 'deny') {
      $this->queryDenyEmpty($values);
    }
    // If allowed, any content that doesn't have a target id should not be
    // restricted.
    else {
      $this->queryAllowEmpty($values);
    }
  }

  /**
   * Perform query when empty values are not allowed.
   *
   * @param array $values
   *   Array of values.
   */
  public function queryDenyEmpty(array $values) {
    // If values are empty then add a 0 because that will never be matched
    // and it will hide it from results.
    if (empty($values)) {
      $values = [0];
      $this->query->condition($this->realField, $values, 'IN');
    }
    else {
      $this->queryCallback($values, $this->query);
    }
  }

  /**
   * Perform query when empty values are allowed.
   *
   * @param array $values
   *   Array of values.
   */
  public function queryAllowEmpty(array $values) {
    $orGroup = $this->query->orConditionGroup();
    $orGroup->condition($this->realField, NULL, 'IS NULL');
    if (!empty($values)) {
      $this->queryCallback($values, $orGroup);
    }

    $this->query->condition($orGroup);
  }

  /**
   * Query method callback.
   *
   * @param array $values
   *   The array of values.
   * @param \Drupal\access_policy\QueryInterface|\Drupal\Core\Database\Query\ConditionInterface $query
   *   The condition/query object. This is set to NULL because the query could
   *   happen to a condition group or to the main query, they do not implement
   *   the same interface.
   */
  public function queryCallback(array $values, $query = NULL) {
    $operator = $this->getOperator();
    $info = $this->operators();
    if (!empty($info[$operator]['method'])) {
      $this->{$info[$operator]['method']}($values, $query);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function ensureMyTable() {
    $base_field_placeholder = $this->query->getBaseTable() . '.' . $this->query->getBaseField();
    $this->query->leftJoin($this->tableAlias, $this->tableAlias, $this->tableAlias . ".entity_id = " . $base_field_placeholder);
  }

  /**
   * The default operators.
   *
   * @return array
   *   Array of operators and their methods.
   */
  public function operators() {
    return [
      'in' => [
        'method' => 'opIn',
      ],
      'not in' => [
        'method' => 'opNotIn',
      ],
    ];
  }

  /**
   * IN operator.
   *
   * @param array $values
   *   The current values.
   * @param \Drupal\access_policy\QueryInterface|\Drupal\Core\Database\Query\ConditionInterface $query
   *   The query or condition object.
   */
  public function opIn(array $values, $query) {
    $query->condition($this->realField, $values, 'IN');
  }

  /**
   * NOT IN operator.
   *
   * @param array $values
   *   The current values.
   * @param \Drupal\access_policy\QueryInterface|\Drupal\Core\Database\Query\ConditionInterface $query
   *   The query or condition object.
   */
  public function opNotIn(array $values, $query) {
    $query->condition($this->realField, $values, 'NOT IN');
  }

}
