<?php

namespace Drupal\access_policy;

use Drupal\Core\Entity\EntityInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;

/**
 * Service for handling Access Policy database content.
 *
 * @package Drupal\access_policy
 */
class ContentAccessPolicyManager {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Access policy information.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * Constructs a ContentAccessPolicyManager object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\access_policy\AccessPolicyInformation $policy_information
   *   Access policy information.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AccessPolicyInformation $policy_information) {
    $this->entityTypeManager = $entity_type_manager;
    $this->accessPolicyInformation = $policy_information;
  }

  /**
   * Set the Access Policy on an entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   * @param array $access_rules
   *   Array of access rules. Note, this currently isn't being used.
   *
   * @throws \Drupal\facets\Exception\Exception
   */
  public function set(EntityInterface $entity, AccessPolicyInterface $access_policy, array $access_rules = []) {
    $this->validateAccessPolicy($access_policy, $entity);
    $entity->set('access_policy', $access_policy->id());
  }

  /**
   * Assign an Access Policy to a content entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The Access Policy.
   *
   * @throws \Exception
   */
  public function assign(EntityInterface $entity, AccessPolicyInterface $access_policy) {
    $this->set($entity, $access_policy);
    $entity->save();
  }

  /**
   * Determine whether the access policy is empty for this entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return bool
   *   TRUE if the access policy is empty; FALSE otherwise.
   */
  public function isEmpty(EntityInterface $entity) {
    return $entity->get('access_policy')->isEmpty();
  }

  /**
   * Remove an Access Policy from a content entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   */
  public function remove(EntityInterface $entity) {
    $entity->set('access_policy', '');
    $entity->save();
  }

  /**
   * Retrieve the Access Policy assigned to a content entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return \Drupal\access_policy\Entity\AccessPolicyInterface
   *   The assigned Access Policy.
   */
  public function getAccessPolicy(EntityInterface $entity) {
    if ($entity->hasField('access_policy') && !$entity->get('access_policy')->isEmpty()) {
      $value = $entity->get('access_policy')->value;
      return $this->entityTypeManager->getStorage('access_policy')->load($value);
    }
    return FALSE;
  }

  /**
   * Validate the access policy.
   *
   * Don't allow entities to be assigned access policies that are not enabled
   * for that entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $policy
   *   The access policy.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @throws \Drupal\facets\Exception\Exception
   */
  private function validateAccessPolicy(AccessPolicyInterface $policy, EntityInterface $entity) {
    $enabled = $this->accessPolicyInformation->isEnabledForEntitiesOfEntityType($policy, $entity->getEntityType(), $entity->bundle());
    if (!$enabled) {
      throw new \Exception(sprintf('The "%1$s" Access Policy is not enabled for "%2$s" bundle "%3$s".', $policy->id(), $entity->getEntityTypeId(), $entity->bundle()));
    }
  }

}
