<?php

namespace Drupal\access_policy\Entity;

use Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface;
use Drupal\Core\Config\Entity\ConfigEntityBase;

/**
 * Represents an AccessPolicy.
 *
 * @ConfigEntityType(
 *   id = "access_policy",
 *   label = @Translation("Access Policy"),
 *   label_collection = @Translation("Access policies"),
 *   label_singular = @Translation("access policy"),
 *   label_plural = @Translation("access policies"),
 *   label_count = @PluralTranslation(
 *     singular = "@count access policy",
 *     plural = "@count access policies",
 *   ),
 *   handlers = {
 *     "list_builder" = "Drupal\access_policy\AccessPolicyListBuilder",
 *     "storage" = "Drupal\access_policy\AccessPolicyStorage",
 *     "access" = "Drupal\access_policy\AccessPolicyAccessControlHandler",
 *     "form" = {
 *       "add" = "Drupal\access_policy\Form\AccessPolicyCreateForm",
 *       "edit" = "Drupal\access_policy\Form\AccessPolicyEditForm",
 *       "delete" = "Drupal\access_policy\Form\AccessPolicyDeleteForm"
 *     },
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider"
 *     },
 *   },
 *   config_prefix = "access_policy",
 *   admin_permission = "administer access policy entities",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "weight" = "weight",
 *   },
 *   links = {
 *     "canonical" = "/admin/people/access_policy/{access_policy}",
 *     "add-form" = "/admin/people/access_policy/add",
 *     "edit-form" = "/admin/people/access_policy/{access_policy}",
 *     "delete-form" = "/admin/people/access_policy/{access_policy}/delete",
 *     "collection" = "/admin/people/access_policy"
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "description",
 *     "weight",
 *     "access_rules",
 *     "access_rule_operator",
 *     "query",
 *     "http_403_response",
 *     "entity_type_auto_select",
 *     "target_entity_type_id",
 *     "target_bundles",
 *     "type",
 *   }
 * )
 */
class AccessPolicy extends ConfigEntityBase implements AccessPolicyInterface {

  /**
   * Unique machine name of the access policy.
   *
   * @var string
   */
  protected $id;

  /**
   * The access policy label.
   *
   * @var string
   */
  protected $label;

  /**
   * The access policy description.
   *
   * @var string
   */
  protected $description;

  /**
   * Weight used to determine the order of access policies.
   *
   * @var int
   */
  protected $weight = 0;

  /**
   * The enabled access rule plugins.
   *
   * @var array
   */
  protected $access_rules = [];

  /**
   * The access rule validation operator.
   *
   * @var string
   */
  protected $access_rule_operator = 'AND';

  /**
   * Flag whether this policy should apply to listing pages.
   *
   * @var bool
   */
  protected $query = TRUE;

  /**
   * The access denied behavior.
   *
   * @var array
   */
  protected $http_403_response = [];

  /**
   * Enable automatic entity type selection based on other policies.
   *
   * @var bool
   */
  protected $entity_type_auto_select = FALSE;

  /**
   * The target bundle.
   *
   * @var string
   */
  protected $target_entity_type_id;

  /**
   * The target bundles.
   *
   * @var array
   */
  protected $target_bundles = [];

  /**
   * The access policy validation type.
   *
   * @var string
   */
  protected $type = 'group';

  /**
   * {@inheritdoc}
   */
  public function label() {
    return $this->label;
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    return $this->description;
  }

  /**
   * {@inheritdoc}
   */
  public function setDescription($description) {
    $this->description = $description;
  }

  /**
   * {@inheritdoc}
   */
  public function getAccessRules() {
    return $this->access_rules;
  }

  /**
   * {@inheritdoc}
   */
  public function getAccessRule($id) {
    return $this->access_rules[$id];
  }

  /**
   * {@inheritdoc}
   */
  public function updateAccessRule($id, array $settings) {
    $this->access_rules[$id]['settings'] = $settings;
  }

  /**
   * {@inheritdoc}
   */
  public function setAccessRules(array $plugins) {
    $this->access_rules = $plugins;
  }

  /**
   * {@inheritdoc}
   */
  public function addAccessRule(AccessRulePluginInterface $access_rule) {
    $definition = $access_rule->getDefinition();
    $name = $this->createAccessRuleUniqueId($definition->getId(), $definition->getPluginId());
    $settings = ['query' => TRUE];
    if (!empty($access_rule->getSettings())) {
      $settings = $access_rule->getSettings();
    }
    $this->access_rules[$name] = [
      'id' => $name,
      'group' => $definition->getGroup(),
      'plugin_id' => $definition->getPluginId(),
      'field' => $definition->getFieldName(),
      'entity_type' => $definition->getEntityType(),
      'settings' => $settings,
      'required' => $definition->isRequired(),
    ];

    return $name;
  }

  /**
   * {@inheritdoc}
   */
  public function removeAccessRule($id) {
    unset($this->access_rules[$id]);
  }

  /**
   * {@inheritdoc}
   */
  public function getAccessRuleOperator() {
    return $this->access_rule_operator;
  }

  /**
   * {@inheritdoc}
   */
  public function setAccessRuleOperator($operator) {
    $this->access_rule_operator = $operator;
  }

  /**
   * Generate a new unique id for the access rule plugin.
   *
   * If the plugin already exists it will append a number to the id.
   * For example: plugin_name_1, plugin_name_2 etc.
   *
   * @param string $id
   *   The original id. This correlates with the id defined in
   *   hook_access_rule_data().
   * @param string $plugin_id
   *   The plugin id.
   *
   * @return string
   *   The access rule unique id.
   */
  private function createAccessRuleUniqueId($id, $plugin_id) {
    $access_rules = $this->access_rules;
    $existing = array_filter($access_rules, function ($rule) use ($plugin_id) {
      if ($rule['plugin_id'] == $plugin_id) {
        return TRUE;
      }
      return FALSE;
    });

    // If any currently exist then append a count to the name.
    if (!empty($existing)) {
      return $id . '_' . count($existing);
    }

    return $id;
  }

  /**
   * {@inheritdoc}
   */
  public function getWeight() {
    return $this->get('weight');
  }

  /**
   * {@inheritdoc}
   */
  public function setWeight($weight) {
    $this->set('weight', $weight);
  }

  /**
   * {@inheritdoc}
   */
  public function isQueryEnabled() {
    return $this->query;
  }

  /**
   * {@inheritdoc}
   */
  public function setQueryEnabled($query = TRUE) {
    $this->query = $query;
  }

  /**
   * {@inheritdoc}
   */
  public function isDefaultHttp403Response() {
    if (empty($this->http_403_response['plugin_id'])) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function getHttp403Response() {
    if (!empty($this->http_403_response['plugin_id'])) {
      $plugin_id = $this->http_403_response['plugin_id'];
      $settings = $this->http_403_response['settings'] ?? [];
      return \Drupal::service('plugin.manager.http_403_response')->createInstance($plugin_id, $settings);
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function setHttp403Response($id, array $settings = []) {
    $this->http_403_response = [
      'plugin_id' => $id,
      'settings' => $settings,
    ];
  }

  /**
   * Set the http 403 response to the system default.
   */
  public function resetHttp403Response() {
    $this->http_403_response = [];
  }

  /**
   * {@inheritdoc}
   */
  public function setAutomaticEntityTypeSelection($auto = TRUE) {
    return $this->entity_type_auto_select = $auto;
  }

  /**
   * {@inheritdoc}
   */
  public function isAutomaticEntityTypeSelection() {
    return $this->entity_type_auto_select;
  }

  /**
   * {@inheritdoc}
   */
  public function getTargetEntityTypeId() {
    return $this->target_entity_type_id;
  }

  /**
   * {@inheritdoc}
   */
  public function getTargetEntityType() {
    return \Drupal::service('entity_type.manager')->getDefinition($this->target_entity_type_id);
  }

  /**
   * {@inheritdoc}
   */
  public function setTargetEntityTypeId($entity_type_id) {
    $this->target_entity_type_id = $entity_type_id;
  }

  /**
   * {@inheritdoc}
   */
  public function getTargetBundles() {
    return $this->target_bundles;
  }

  /**
   * {@inheritdoc}
   */
  public function setTargetBundles(array $bundles) {
    $this->target_bundles = $bundles;
  }

  /**
   * Set the validation type.
   *
   * @param string $type
   *   The validation type plugin id.
   */
  public function setType($type) {
    $this->type = $type;
  }

  /**
   * Get the validation type plugin name.
   *
   * @return string
   *   The validation type plugin id.
   */
  public function getType() {
    return $this->type;
  }

  /**
   * Get the access policy type plugin.
   *
   * @return \Drupal\access_policy\Plugin\AccessPolicyType\AccessPolicyTypeInterface
   *   The access policy type interface.
   */
  public function getTypePlugin() {
    return \Drupal::service('plugin.manager.access_policy_type')->createInstance($this->type);
  }

}
