<?php

namespace Drupal\access_policy\Entity;

use Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface;
use Drupal\Core\Config\Entity\ConfigEntityInterface;

/**
 * Provides an interface defining an AccessPolicy entity.
 *
 * @package Drupal\access_policy\Entity
 */
interface AccessPolicyInterface extends ConfigEntityInterface {

  /**
   * Returns label value.
   *
   * @return string
   *   The label value.
   */
  public function label();

  /**
   * Gets the access policy description.
   *
   * @return string
   *   The description.
   */
  public function getDescription();

  /**
   * Sets the access policy description.
   *
   * @param string $description
   *   The description.
   */
  public function setDescription($description);

  /**
   * Get access rule plugins.
   *
   * @return array
   *   The access rule plugins.
   */
  public function getAccessRules();

  /**
   * Get access rule plugin by id.
   *
   * @param string $id
   *   The access rule id. Note that this shouldn't be confused with plugin_id.
   *   This is the unique id that gets generated when added to a policy.
   *
   * @return array
   *   The access rule plugin.
   */
  public function getAccessRule($id);

  /**
   * Update an access rule's settings.
   *
   * @param string $id
   *   The unique id of the access rule.
   * @param array $settings
   *   The access rule settings.
   */
  public function updateAccessRule($id, array $settings);

  /**
   * Set access rule plugins.
   *
   * @param array $plugins
   *   The access rule plugins.
   */
  public function setAccessRules(array $plugins);

  /**
   * Add an access rule plugin.
   *
   * @param \Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface $access_rule
   *   The full instantiated access rule plugin.
   *
   * @return string
   *   The new unique access rule id.
   */
  public function addAccessRule(AccessRulePluginInterface $access_rule);

  /**
   * Remove an access rule plugin.
   *
   * @param string $id
   *   The unique id of the plugin.
   */
  public function removeAccessRule($id);

  /**
   * Get the access rule operator.
   *
   * @return string
   *   The validation operator. Default value is AND.
   */
  public function getAccessRuleOperator();

  /**
   * Set the access rule operator.
   *
   * @param string $operator
   *   The operator. Accepted values are AND and OR.
   */
  public function setAccessRuleOperator($operator);

  /**
   * Get the weight.
   *
   * @return int
   *   The weight.
   */
  public function getWeight();

  /**
   * Set the weight.
   *
   * @param int $weight
   *   The weight.
   */
  public function setWeight(int $weight);

  /**
   * Determine whether the policy should be applied to queries.
   *
   * @return bool
   *   TRUE if the policy is enabled; FALSE otherwise.
   */
  public function isQueryEnabled();

  /**
   * Set the policy as enabled or disabled for queries.
   *
   * In some cases you may only want to restrict the access policy to view,
   * edit, delete etc. This allows you to disable the policy on query pages
   * entirely.
   *
   * @param bool $query
   *   TRUE if the policy should be enabled; FALSE otherwise.
   */
  public function setQueryEnabled($query = TRUE);

  /**
   * Determine whether the http 403 response is the system default.
   *
   * @return bool
   *   TRUE if the response is default; FALSE otherwise.
   */
  public function isDefaultHttp403Response();

  /**
   * Get the access denied behavior plugin.
   *
   * @return \Drupal\access_policy\Plugin\Http403Response\Http403ResponseInterface
   *   The access denied plugin or FALSE if not set.
   */
  public function getHttp403Response();

  /**
   * Set the access denied behavior plugin and settings.
   *
   * @param string $id
   *   The access denied behavior plugin id.
   * @param array $settings
   *   The plugin settings.
   */
  public function setHttp403Response($id, array $settings = []);

  /**
   * Set the http 403 response to the system default.
   */
  public function resetHttp403Response();

  /**
   * Set the entity type selection to automatic.
   *
   * @param bool $auto
   *   TRUE sets the selection to automatic; FALSE sets it to manual (default).
   */
  public function setAutomaticEntityTypeSelection($auto = TRUE);

  /**
   * Determine whether the entity type selection is automatic or not.
   *
   * @return bool
   *   TRUE if entity type selection is automaic; FALSE otherwise.
   */
  public function isAutomaticEntityTypeSelection();

  /**
   * Get the target entity type id.
   *
   * @return string
   *   The target entity type id
   */
  public function getTargetEntityTypeId();

  /**
   * Get the target entity type.
   *
   * @return \Drupal\Core\Entity\EntityTypeInterface
   *   The target entity type.
   */
  public function getTargetEntityType();

  /**
   * Set the target entity type id.
   *
   * @param string $entity_type_id
   *   The entity type id.
   */
  public function setTargetEntityTypeId(string $entity_type_id);

  /**
   * Get the bundles configured with this policy.
   *
   * @return array
   *   Array of bundles.
   */
  public function getTargetBundles();

  /**
   * Set the bundles that should be supported with this policy.
   *
   * @param array $bundles
   *   Array of bundles.
   */
  public function setTargetBundles(array $bundles);

}
