<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Entity\ContentAccessPolicy;
use Drupal\access_policy\Entity\ContentAccessPolicyInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\views\Views;

/**
 * Defines a class for reacting to entity events.
 *
 * @internal
 */
class EntityOperations implements ContainerInjectionInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity storage handler.
   *
   * @var \Drupal\access_policy\ContentAccessPolicyManager
   */
  protected $storage;

  /**
   * The access policy selection service.
   *
   * @var \Drupal\access_policy\AccessPolicySelectionInterface
   */
  protected $selection;

  /**
   * The access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $information;

  /**
   * Constructs a new EntityOperations object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\ContentAccessPolicyManager $storage
   *   The access policy content entity handler.
   * @param \Drupal\access_policy\AccessPolicySelectionInterface $selection
   *   The access policy selection service.
   * @param \Drupal\access_policy\AccessPolicyInformation $information
   *   The access policy information service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, ContentAccessPolicyManager $storage, AccessPolicySelectionInterface $selection, AccessPolicyInformation $information) {
    $this->entityTypeManager = $entity_type_manager;
    $this->storage = $storage;
    $this->selection = $selection;
    $this->information = $information;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('access_policy.content_policy_manager'),
      $container->get('access_policy.selection'),
      $container->get('access_policy.information'),
    );
  }

  /**
   * Assign a default access policy to a new entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The new entity.
   *
   * @see hook_entity_presave()
   */
  public function entityPresave(EntityInterface $entity) {
    if ($this->information->hasEnabledForEntitiesOfEntityType($entity->getEntityType(), $entity->bundle()) && $entity->isNew()) {
      if (!$this->storage->getAccessPolicy($entity)) {
        $default = $this->selection->getDefaultPolicy($entity, \Drupal::currentUser());
        if (!empty($default)) {
          $this->storage->set($entity, $default);
        }
      }
    }
  }

  /**
   * Setting the default access policy on the entity form.
   *
   * This is primarily for supporting constraint validation during initial
   * creation of the entity. The access policy actually gets saved as part of
   * hook_entity_presave.
   *
   * @see hook_form_alter();
   */
  public function entityFormAlter(&$form, FormStateInterface $form_state, $form_id) {
    $form_obj = $form_state->getFormObject();
    $entity = $form_obj->getEntity();
    if ($this->information->hasEnabledForEntitiesOfEntityType($entity->getEntityType(), $entity->bundle()) && $entity->isNew()) {
      $default = $this->selection->getDefaultPolicy($entity, \Drupal::currentUser());
      if (!empty($default)) {
        $this->storage->set($entity, $default);
      }
    }
  }

  /**
   * Entity insert hook.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity that was just saved.
   *
   * @see hook_entity_insert()
   */
  public function entityInsert(EntityInterface $entity) {
    if ($this->information->isAccessControlledEntity($entity)) {
      $this->updateOrCreateFromEntity($entity);
    }
  }

  /**
   * Entity update hook.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity that was just saved.
   *
   * @see hook_entity_update()
   */
  public function entityUpdate(EntityInterface $entity) {
    if ($this->information->isAccessControlledEntity($entity)) {
      $this->updateOrCreateFromEntity($entity);
    }
  }

  /**
   * Creates or updates the access policy of an entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to update or create an access policy for.
   */
  protected function updateOrCreateFromEntity(EntityInterface $entity) {
    /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
    $content_access_policy = ContentAccessPolicy::loadFromAccessControlledEntity($entity);
    /** @var \Drupal\Core\Entity\ContentEntityStorageInterface $storage */
    $storage = $this->entityTypeManager->getStorage('content_access_policy');

    if (!($content_access_policy instanceof ContentAccessPolicyInterface)) {
      $content_access_policy = $storage->create([
        'content_entity_type_id' => $entity->getEntityTypeId(),
        'content_entity_id' => $entity->id(),
      ]);
    }

    // Create the ContentAccessPolicy entity for the inserted entity.
    $access_policy = $entity->access_policy->value;

    $content_access_policy->set('access_policy', $access_policy);
    ContentAccessPolicy::updateOrCreateFromEntity($content_access_policy);
  }

  /**
   * Entity delete hook.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being deleted.
   *
   * @see hook_entity_delete()
   */
  public function entityDelete(EntityInterface $entity) {
    $content_access_policy = ContentAccessPolicy::loadFromAccessControlledEntity($entity);
    if ($content_access_policy) {
      $content_access_policy->delete();
    }
  }

  /**
   * Update all access policies with automatic entity type selection.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The access policy entity.
   */
  public function accessPolicyUpdate(EntityInterface $entity) {
    $bundles = $this->getAllBundles();
    $auto_policies = $this->entityTypeManager->getStorage('access_policy')->loadByProperties(['entity_type_auto_select' => TRUE]);
    if (!empty($auto_policies)) {
      foreach ($auto_policies as $policy) {
        $current_bundles = $policy->getTargetBundles();
        if ($current_bundles !== $bundles) {
          $policy->setTargetBundles($bundles);
          $policy->save();
        }
      }
    }

    // Clear field cache so extra field is added or removed.
    \Drupal::service('entity_field.manager')->clearCachedFieldDefinitions();
    // Clear the views data cache so the extra field is available in views.
    if (\Drupal::moduleHandler()->moduleExists('views')) {
      Views::viewsData()->clear();
    }
  }

  /**
   * Rebuild the routes.
   *
   * If this is the first access policy for a given entity type we need to
   * rebuild the routes so that the "Access" tab appears.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The access policy entity.
   */
  public function accessPolicyInsert(EntityInterface $entity) {
    $existing_policies = $this->entityTypeManager->getStorage('access_policy')->loadByProperties(['target_entity_type_id' => $entity->getTargetEntityTypeId()]);
    // If this is the first policy for a given entity type then clear the
    // routes and register the access rule data.
    if (count($existing_policies) == 1) {
      \Drupal::service('router.builder')->setRebuildNeeded();
      \Drupal::service('access_policy.access_rule_data')->clear();
    }

    // Clear field cache so extra field is added or removed.
    \Drupal::service('entity_field.manager')->clearCachedFieldDefinitions();
    // Clear the views data cache so the extra field is available in views.
    if (\Drupal::moduleHandler()->moduleExists('views')) {
      Views::viewsData()->clear();
    }
  }

  /**
   * Get all the entity types and bundles across policies.
   *
   * @return array
   *   The array of entity types and bundles.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getAllBundles() {
    $policies = $this->entityTypeManager->getStorage('access_policy')->loadByProperties(['entity_type_auto_select' => FALSE]);
    $merged_bundles = [];
    foreach ($policies as $policy) {
      $bundles = $policy->getTargetBundles();
      $merged_bundles = array_unique(array_merge($bundles, $merged_bundles));
    }

    return $merged_bundles;
  }

}
