<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Plugin\Field\AccessPolicyFieldItemList;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Manipulates entity type information.
 *
 * This class contains primarily bridged hooks for compile-time or
 * cache-clear-time hooks. Runtime hooks should be placed in EntityOperations.
 *
 * @internal
 */
class EntityTypeInfo {

  use StringTranslationTrait;

  /**
   * Provides custom base field definitions for a content entity type.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   An array of field definitions, keyed by field name.
   *
   * @see hook_entity_base_field_info()
   */
  public function entityBaseFieldInfo(EntityTypeInterface $entity_type) {
    if (!\Drupal::service('access_policy.information')->isAccessControlledEntityType($entity_type)) {
      return [];
    }

    $fields = [];
    $fields['access_policy'] = BaseFieldDefinition::create('string')
      ->setLabel($this->t('Access policy'))
      ->setDescription($this->t('The access policy of this piece of content.'))
      ->setComputed(TRUE)
      ->setClass(AccessPolicyFieldItemList::class)
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE)
      ->setReadOnly(FALSE)
      ->setRevisionable(FALSE)
      ->setTranslatable(FALSE);

    return $fields;
  }

  /**
   * Alter entity types and access policy access form.
   *
   * @param array $entity_types
   *   Array of entity types.
   *
   * @see hook_entity_type_alter()
   */
  public function entityTypeAlter(array &$entity_types) {
    foreach ($entity_types as $entity_type) {
      if (\Drupal::service('access_policy.information')->isSupportedEntityType($entity_type)) {
        $path = $entity_type->getLinkTemplate('canonical');
        $entity_type->setLinkTemplate('access-policy-form', $path . '/access');
        $entity_type->setFormClass('assign access policy', 'Drupal\access_policy\Form\AccessPolicyEntityForm');
      }
    }
  }

}
