<?php

namespace Drupal\access_policy\EventSubscriber;

use Drupal\access_policy\ContentAccessPolicyManager;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\EventSubscriber\DefaultExceptionHtmlSubscriber;
use Drupal\Core\Routing\RedirectDestinationInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;
use Symfony\Component\HttpKernel\HttpKernelInterface;
use Symfony\Component\Routing\Matcher\UrlMatcherInterface;

/**
 * Exception subscriber for handling access policy 403 HTML error pages.
 */
class AccessPolicyExceptionHtmlSubscriber extends DefaultExceptionHtmlSubscriber {

  /**
   * The content policy manager service.
   *
   * @var \Drupal\access_policy\ContentAccessPolicyManager
   */
  protected $contentPolicyManager;

  /**
   * Constructs a new AccessPolicyExceptionHtmlSubscriber.
   *
   * @param \Symfony\Component\HttpKernel\HttpKernelInterface $http_kernel
   *   The HTTP Kernel service.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger service.
   * @param \Drupal\Core\Routing\RedirectDestinationInterface $redirect_destination
   *   The redirect destination service.
   * @param \Symfony\Component\Routing\Matcher\UrlMatcherInterface $access_unaware_router
   *   A router implementation which does not check access.
   * @param \Drupal\access_policy\ContentAccessPolicyManager $content_policy_manager
   *   The content policy manager service.
   */
  public function __construct(HttpKernelInterface $http_kernel, LoggerInterface $logger, RedirectDestinationInterface $redirect_destination, UrlMatcherInterface $access_unaware_router, ContentAccessPolicyManager $content_policy_manager) {
    parent::__construct($http_kernel, $logger, $redirect_destination, $access_unaware_router);
    $this->contentPolicyManager = $content_policy_manager;
  }

  /**
   * {@inheritdoc}
   */
  protected static function getPriority() {
    return -25;
  }

  /**
   * {@inheritdoc}
   */
  public function on403(ExceptionEvent $event) {
    $entity = $this->getEntity($event);
    if (!empty($entity)) {
      $policy = $this->contentPolicyManager->getAccessPolicy($entity);
      if ($policy && !$policy->isDefaultHttp403Response()) {
        $this->makeSubrequest($event, '/access_policy/403/' . $policy->id(), Response::HTTP_FORBIDDEN);
      }
    }
  }

  /**
   * Get the current entity.
   *
   * @param \Symfony\Component\HttpKernel\Event\ExceptionEvent $event
   *   The exception event.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The entity object if found.
   */
  public function getEntity(ExceptionEvent $event) {
    $params = $event->getRequest()->attributes;
    if ($params->count() > 0) {
      foreach ($params as $parameter) {
        if ($parameter instanceof ContentEntityInterface) {
          return $parameter;
        }
      }
    }
  }

}
