<?php

namespace Drupal\access_policy\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;

/**
 * The form for adding access rules to an access policy.
 *
 * This is actually a pseudo form, it's being loaded as part of
 * AccessRuleMultistepForm. Not in the route.
 *
 * @internal
 */
class AccessPolicyAddRuleForm extends AccessRuleFormBase {

  /**
   * The access policy entity object.
   *
   * @var \Drupal\access_policy\Entity\AccessPolicyInterface
   */
  protected $accessPolicy;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'access_policy_add_rule_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, AccessPolicyInterface $access_policy = NULL) {
    $this->accessPolicy = $access_policy;
    $form['#attributes']['class'][] = 'access-rule-modal-form';

    $form['filter'] = [
      '#title' => $this->t('Search'),
      '#type' => 'textfield',
      '#ajax' => [
        'event' => 'change',
        'keypress' => TRUE,
        'callback' => [$this, 'ajaxFilterCallback'],
      ],
    ];

    // Show access rule plugins (if available).
    $definitions = $this->getApplicableAccessRules();

    // Sort the access rules by label.
    $definitions = $this->sortAccessRules($definitions);

    // Filter the access rules if applicable.
    if (!empty($form_state->getValue('filter'))) {
      $definitions = $this->filterAccessRules($definitions, $form_state->getValue('filter'));
    }

    $options = [];
    foreach ($definitions as $id => $access_rule) {
      $options[$id] = [
        'title' => $access_rule->getDefinition()->getLabel(),
        'description' => $access_rule->getDefinition()->getDescription(),
      ];
    }

    $form['access_rules'] = [
      '#type' => 'tableselect',
      '#title' => $this->t('Access rule'),
      '#js_select' => FALSE,
      '#header' => [
        'title' => $this->t('Title'),
        'description' => $this->t('Description'),
      ],
      '#element_validate' => [[static::class, 'addRuleValidate']],
      '#options' => $options,
      '#empty' => $this->t('No access rules found.'),
    ];

    return $form;
  }

  /**
   * Validate the access rule selection.
   *
   * @param array $form
   *   The current form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public static function addRuleValidate(array $form, FormStateInterface $form_state) {
    $access_rules = $form_state->getValue('access_rules');
    $selected_rules = array_filter($access_rules);
    if (empty($selected_rules)) {
      $form_state->setErrorByName('access_rules', 'Please select an access rule.');
    }
    if (count($selected_rules) > 1) {
      $form_state->setErrorByName('access_rules', 'Only one access rule can be added at a time.');
    }
  }

  /**
   * Ajax callback to reload the form filtered access rules.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   An ajax response object.
   */
  public function ajaxFilterCallback(array $form, FormStateInterface $form_state) {
    $response = new AjaxResponse();
    $response->addCommand(new ReplaceCommand('.access-rule-modal-form', $form));
    return $response;
  }

  /**
   * Filter access rules.
   *
   * @param array $access_rules
   *   Array of access rule plugins.
   * @param string $search
   *   The search string.
   *
   * @return array
   *   Array of filtered definitions.
   */
  public function filterAccessRules(array $access_rules, $search) {
    return array_filter($access_rules, function ($rule) use ($search) {
      if (stripos($rule->getDefinition()->getLabel(), $search) !== FALSE) {
        return TRUE;
      }
      return FALSE;
    });
  }

  /**
   * Sort the access rules by label.
   *
   * @param array $access_rules
   *   Array of access rule plugins.
   *
   * @return array
   *   Array of sorted access rule plugins.
   */
  protected function sortAccessRules(array $access_rules) {
    $key_value = [];
    foreach ($access_rules as $id => $rule) {
      $key_value[$id] = $rule->getDefinition()->getLabel();
    }
    natsort($key_value);

    $sorted = [];
    foreach ($key_value as $id => $label) {
      $sorted[$id] = $access_rules[$id];
    }

    return $sorted;
  }

  /**
   * Returns all applicable access rules for this policy.
   *
   * @return array
   *   Array of applicable access rules.
   */
  public function getApplicableAccessRules() {
    $supported_groups = [
      'global',
      'user',
      $this->accessPolicy->getTargetEntityTypeId(),
    ];

    $rules = [];
    foreach ($supported_groups as $group) {
      $group_rules = $this->accessRuleManager()->getAccessRulePluginsByGroup($group);
      $rules = array_merge($rules, $group_rules);
    }
    return $rules;
  }

}
