<?php

namespace Drupal\access_policy\Form;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\CloseDialogCommand;
use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The form for editing entity types associated with an access policy.
 *
 * @internal
 */
class AccessPolicyConfigureBundlesForm extends FormBase {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity type bundle information service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $bundleInfo;

  /**
   * The access policy entity object.
   *
   * @var \Drupal\access_policy\Entity\AccessPolicyInterface
   */
  protected $accessPolicy;

  /**
   * The entity type id.
   *
   * @var string
   */
  protected $entityTypeId;

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfoInterface $bundle_info) {
    $this->entityTypeManager = $entity_type_manager;
    $this->bundleInfo = $bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'access_policy_entity_type_edit_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, AccessPolicyInterface $access_policy = NULL) {
    $this->accessPolicy = $access_policy;
    $this->entityTypeId = $access_policy->getTargetEntityTypeId();
    $entity_type = $this->entityTypeManager->getDefinition($this->entityTypeId);

    $options = $defaults = [];
    foreach ($this->bundleInfo->getBundleInfo($this->entityTypeId) as $bundle_id => $bundle) {
      $is_enabled = $this->isSelectedBundle($bundle_id);
      $options[$bundle_id] = [
        'title' => ['data' => ['#title' => $bundle['label']]],
        'type' => $bundle['label'],
      ];
      $defaults[$bundle_id] = $is_enabled;
    }

    if (!empty($options)) {
      $bundles_header = $this->t('All @entity_type types', ['@entity_type' => $entity_type->getLabel()]);
      if ($bundle_entity_type_id = $entity_type->getBundleEntityType()) {
        $bundles_header = $this->t('All @entity_type_plural_label', ['@entity_type_plural_label' => $this->entityTypeManager->getDefinition($bundle_entity_type_id)->getPluralLabel()]);
      }
      $form['bundles'] = [
        '#type' => 'tableselect',
        '#header' => [
          'type' => $bundles_header,
        ],
        '#options' => $options,
        '#default_value' => $defaults,
        '#attributes' => ['class' => ['no-highlight']],
      ];
    }

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#button_type' => 'primary',
      '#value' => $this->t('Save'),
      '#ajax' => [
        'callback' => [$this, 'ajaxcallback'],
      ],
    ];
    $form['actions']['cancel'] = [
      '#type' => 'button',
      '#value' => $this->t('Cancel'),
      '#ajax' => [
        'callback' => [$this, 'ajaxcallback'],
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $selected_bundles = $this->accessPolicy->getTargetBundles();

    // Reset the bundles.
    $selected_bundles = [];
    foreach ($form_state->getValue('bundles') as $bundle_id => $checked) {
      if ($checked) {
        $selected_bundles[] = $bundle_id;
      }
    }

    $this->accessPolicy->setTargetBundles($selected_bundles);
    $this->accessPolicy->save();
  }

  /**
   * Determine whether a bundle is enabled for this entity type.
   *
   * @param string $bundle
   *   The name of the bundle.
   *
   * @return bool
   *   TRUE if it is enabled; FALSE otherwise.
   */
  public function isSelectedBundle($bundle) {
    $bundles = $this->accessPolicy->getTargetBundles();
    if (in_array($bundle, $bundles)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Load the selected bundles.
   *
   * @param string $entity_type_id
   *   The entity type id.
   *
   * @return array
   *   Array of selected entities.
   */
  public function getSelectedBundles($entity_type_id) {
    $selected_bundles = [];
    $bundle_info = $this->bundleInfo->getBundleInfo($entity_type_id);
    foreach ($bundle_info as $id => $bundle) {
      if ($this->isSelectedBundle($id)) {
        $selected_bundles[$id] = $bundle['label'];
      }
    }

    return $selected_bundles;
  }

  /**
   * Ajax callback to close the modal and update the selected bundles.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   An ajax response object.
   */
  public function ajaxCallback() {
    $selected_bundles = $this->getSelectedBundles($this->entityTypeId);

    $selected_bundles_list = [
      '#theme' => 'item_list',
      '#items' => $selected_bundles,
      '#context' => ['list_style' => 'comma-list'],
      '#empty' => $this->t('None'),
    ];
    $response = new AjaxResponse();
    $response->addCommand(new CloseDialogCommand());
    $response->addCommand(new HtmlCommand('#selected-bundles', $selected_bundles_list));
    return $response;
  }

}
