<?php

namespace Drupal\access_policy\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;

/**
 * The form for configuring access rules on an access policy.
 *
 * @internal
 */
class AccessPolicyConfigureRuleForm extends AccessRuleFormBase {

  /**
   * The Access Rule id.
   *
   * @var string
   */
  protected $id;

  /**
   * Flag to determine whether we're configuring a new access rule.
   *
   * @var bool
   */
  protected $isNew;

  /**
   * The access policy entity object.
   *
   * @var \Drupal\access_policy\Entity\AccessPolicyInterface
   */
  protected $accessPolicy;

  /**
   * The current access rule plugin.
   *
   * @var \Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface
   */
  protected $accessRulePlugin;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'access_policy_configure_rule_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, AccessPolicyInterface $access_policy = NULL, $id = NULL, $is_new = NULL) {
    $this->accessPolicy = $access_policy;
    $this->id = $id;
    $this->isNew = $is_new;

    // We need to ensure that the plugin id is preserved during validation.
    // This is passed up to the add rule form when we're configuring an access
    // rule for the first time.
    $form['id'] = [
      '#type' => 'hidden',
      '#value' => $id,
    ];

    // Show any validation errors in the modal.
    $form['status_messages'] = [
      '#type' => 'status_messages',
      '#weight' => -10,
    ];

    // This is necessary for modal validation.
    $form['#attributes']['class'][] = 'access-rule-modal-form';

    $plugin_object = $this->getAccessRulePlugin();
    $form += $plugin_object->buildSettingsForm($form, $form_state);

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $form_obj = $form_state->getFormObject();
    $plugin = $this->getAccessRulePlugin();
    $plugin->submitSettingsForm($form, $form_state);

    $this->accessPolicy->updateAccessRule($form_obj->id, $plugin->getSettings());
    $this->accessPolicy->save();
  }

  /**
   * Validate the access rule plugin.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $plugin = $this->getAccessRulePlugin();
    $plugin->validateSettingsForm($form, $form_state);
  }

  /**
   * Get the current access rule plugin.
   *
   * @return \Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface
   *   The access rule plugin.
   */
  protected function getAccessRulePlugin() {
    // If the access rule is new instantiate a new plugin.
    if (empty($this->accessRulePlugin)) {
      if ($this->isNew) {
        [$group, $name] = explode(':', $this->id);
        $this->accessRulePlugin = $this->accessRulePluginManager->getHandler($group, $name, ['access_policy' => $this->accessPolicy->id()]);
      }
      // Otherwise use the existing access rule.
      else {
        $this->accessRulePlugin = $this->accessRulePluginManager->getAccessRulePlugin($this->accessPolicy, $this->id);
      }
    }

    return $this->accessRulePlugin;
  }

  /**
   * Route title callback.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   * @param string $id
   *   The access rule plugin id.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The route title.
   */
  public function getTitle(AccessPolicyInterface $access_policy, $id) {
    $plugin = $this->accessRulePluginManager->getAccessRulePlugin($access_policy, $id);
    return $this->t('Configure access rule: @label', ['@label' => $plugin->getDefinition()->getLabel()]);
  }

}
