<?php

namespace Drupal\access_policy\Form;

use Drupal\access_policy\AccessRuleWidgetPluginManager;
use Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\access_policy\AccessRulePluginManager;
use Drupal\access_policy\ContentAccessPolicyManager;
use Drupal\access_policy\AccessPolicySelectionInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Session\AccountInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form to select an active Access Policy.
 *
 * @internal
 */
class AccessPolicyEntityForm extends EntityForm {

  /**
   * The ContentAccessPolicyManager service.
   *
   * @var \Drupal\access_policy\ContentAccessPolicyManager
   */
  protected $contentAccessPolicyManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The AccessRulePluginManager service.
   *
   * @var \Drupal\access_policy\AccessRulePluginManager
   */
  protected $accessRulePluginManager;

  /**
   * The AccessRuleWidgetPluginManager service.
   *
   * @var \Drupal\access_policy\AccessRuleWidgetPluginManager
   */
  protected $accessRuleWidgetPluginManager;

  /**
   * Access policy selection service.
   *
   * @var \Drupal\access_policy\AccessPolicySelectionInterface
   */
  protected $policySelection;

  /**
   * Constructs a AccessPolicyEntityForm object.
   *
   * @param \Drupal\access_policy\ContentAccessPolicyManager $content_access_policy_manager
   *   The ContentAccessPolicyManager service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\AccessRulePluginManager $access_rule_plugin_manager
   *   The AccessRulePluginManager service.
   * @param \Drupal\access_policy\AccessRuleWidgetPluginManager $widget_plugin_manager
   *   The AccessRuleWidgetPluginManager service.
   * @param \Drupal\access_policy\AccessPolicySelectionInterface $selection
   *   The access policy selection service.
   */
  public function __construct(ContentAccessPolicyManager $content_access_policy_manager, EntityTypeManagerInterface $entity_type_manager, AccessRulePluginManager $access_rule_plugin_manager, AccessRuleWidgetPluginManager $widget_plugin_manager, AccessPolicySelectionInterface $selection) {
    $this->contentAccessPolicyManager = $content_access_policy_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->accessRulePluginManager = $access_rule_plugin_manager;
    $this->accessRuleWidgetPluginManager = $widget_plugin_manager;
    $this->policySelection = $selection;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('access_policy.content_policy_manager'),
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.access_rule'),
      $container->get('plugin.manager.access_rule_widget'),
      $container->get('access_policy.selection'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getBaseFormId() {
    return $this->entity->getEntityTypeId() . '_access_policy_form';
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'access_policy_entity_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $entity_type_id = NULL) {
    $entity = $this->getEntity();

    $form['access_policy_settings'] = [
      '#type' => 'fieldset',
      '#prefix' => '<div id="access-policy-settings">',
      '#suffix' => '</div>',
    ];

    // Only display Access Policies permitted for this Content Entity Type.
    $access_policies = $this->policySelection->getAllowedAccessPolicies($entity, \Drupal::currentUser());

    // Get the submitted submitted access policy.
    $values = $form_state->getValues();
    if (isset($values['access_policy'])) {
      // If a value is provided then load that.
      if (!empty($values['access_policy'])) {
        $current_policy = $this->loadAccessPolicy($values['access_policy']);
      }
      // Otherwise set it to empty.
      else {
        $current_policy = FALSE;
      }
    }
    // Always show the current access policy, even if it's one that is not
    // allowed for that user.
    else {
      $current_policy = $this->getCurrentAccessPolicy($entity);
    }
    if (!empty($current_policy)) {
      $access_policies[] = $current_policy;
    }

    $options = $this->getAccessPolicyOptions($access_policies);

    if (empty($options)) {
      $form['access_policy_settings']['empty'] = $this->buildFormEmpty($form, $form_state);
      return $form;
    }

    $form['access_policy_settings']['access_policy'] = [
      '#type' => 'select',
      '#title' => $this->t('Set access to:'),
      '#default_value' => ($current_policy) ? $current_policy->id() : FALSE,
      '#options' => $options,
      '#ajax' => [
        'callback' => '::loadAccessRuleSettings',
        'wrapper' => 'access-policy-settings',
        'event' => 'change',
      ],
      '#disabled' => !$this->isAllowedOrEmpty($entity, \Drupal::currentUser()),
      '#description' => ($current_policy) ? $current_policy->getDescription() : '',
    ];

    if ($this->policySelection->emptyPoliciesAllowed()) {
      $form['access_policy_settings']['access_policy']['#empty_option'] = $this->t('- Select -');
    }

    // Build form elements defined by Access Rules for selected Access Policy.
    if ($current_policy) {
      $applicable_plugins = $this->accessRulePluginManager->getApplicableAccessRulePlugins($current_policy, $entity);
      foreach ($applicable_plugins as $plugin_name => $plugin) {
        if ($this->isWidgetEnabled($plugin)) {
          $widget = $this->createWidgetInstance($entity, $plugin);

          $form['access_policy_settings'][$plugin_name] = [];
          $subform_state = SubformState::createForSubform($form['access_policy_settings'][$plugin_name], $form, $form_state);
          $form['access_policy_settings'][$plugin_name] = $widget->buildForm($form['access_policy_settings'][$plugin_name], $subform_state);
        }
      }
    }

    $form['actions']['#type'] = 'actions';
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
      '#button_type' => 'primary',
      '#submit' => ['::submitForm', '::save'],
    ];

    return $form;
  }

  /**
   * Determine whether the access rule widget is enabled.
   *
   * @param \Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface $plugin
   *   The access rule plugin.
   *
   * @return bool
   *   TRUE if the access rule widget is enabled; FALSE otherwise.
   */
  private function isWidgetEnabled(AccessRulePluginInterface $plugin) {
    $config = $plugin->getSettings();
    if ($plugin->getDefinition()->getWidget() && $config['widget']['show']) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Create a widget instance for this context.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param \Drupal\access_policy\Plugin\AccessRule\AccessRulePluginInterface $plugin
   *   The access rule plugin.
   *
   * @return \Drupal\access_policy\Plugin\AccessRuleWidget\AccessRuleWidgetPluginInterface
   *   The access rule widget plugin interface.
   */
  private function createWidgetInstance(EntityInterface $entity, AccessRulePluginInterface $plugin) {
    $widget = $this->accessRuleWidgetPluginManager->createInstance($plugin->getDefinition()->getWidget());
    $widget->initialize($plugin->getDefinition());

    $config = $plugin->getSettings();
    $widget->setEntity($entity);
    $settings = $config['widget']['settings'] ?? [];
    $widget->setWidgetSettings($settings);
    return $widget;
  }

  /**
   * Render a message if no access policies are available.
   *
   * @param array $form
   *   The current form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   The form.
   */
  public function buildFormEmpty(array $form, FormStateInterface $form_state) {
    return [
      '#markup' => $this->t('No access policies are currently available.'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $entity = $this->getEntity();
    $access_policy_id = $form_state->getValue('access_policy');
    if (!empty($access_policy_id)) {
      $access_policy = $this->loadAccessPolicy($access_policy_id);

      $plugins = $this->accessRulePluginManager->getApplicableAccessRulePlugins($access_policy, $entity);
      foreach ($plugins as $plugin) {
        if ($this->isWidgetEnabled($plugin)) {
          $widget = $this->createWidgetInstance($entity, $plugin);
          $widget->validateForm($form, $form_state);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $entity = $this->getEntity();
    $access_policy = $form_state->getValue('access_policy');
    if (!empty($access_policy)) {
      $access_policy = $this->loadAccessPolicy($access_policy);

      $plugins = $this->accessRulePluginManager->getApplicableAccessRulePlugins($access_policy, $entity);
      foreach ($plugins as $plugin) {
        if ($this->isWidgetEnabled($plugin)) {
          $widget = $this->createWidgetInstance($entity, $plugin);
          $widget->submitForm($form, $form_state);
        }
      }

      $this->contentAccessPolicyManager->set($entity, $access_policy);
      \Drupal::messenger()->addMessage($this->t('Access has been changed to @policy.', ['@policy' => $access_policy->label()]));
    }
    else {
      $this->contentAccessPolicyManager->remove($entity);
      \Drupal::messenger()->addMessage($this->t('Access policy has been removed.'));
    }

  }

  /**
   * Ajax callback for updating Access Rules.
   *
   * @param array &$form
   *   The built form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The updated content.
   */
  public function loadAccessRuleSettings(array &$form, FormStateInterface $form_state) {
    return $form['access_policy_settings'];
  }

  /**
   * Return Access Policy form options.
   *
   * @param array $entities
   *   An array of Access Policy entities.
   *
   * @return array
   *   The Access Policy form options.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getAccessPolicyOptions(array $entities) {
    $options = [];
    foreach ($entities as $entity) {
      $options[$entity->id()] = $entity->label();
    }
    return $options;
  }

  /**
   * Get the current Access Policy.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return bool|\Drupal\access_policy\Entity\AccessPolicyInterface
   *   The current access policy entity or FALSE if unset.
   */
  public function getCurrentAccessPolicy(EntityInterface $entity) {
    $access_policy = $this->contentAccessPolicyManager->getAccessPolicy($entity);
    if (!empty($access_policy)) {
      return $access_policy;
    }

    if (empty($access_policy) && !$this->contentAccessPolicyManager->isEmpty($entity)) {
      $default_policy = $this->policySelection->getDefaultPolicy($entity, \Drupal::currentUser()) ?? '';
      if ($default_policy) {
        return $default_policy;
      }
    }

    return FALSE;
  }

  /**
   * Determine whether this access policy is allowed or empty.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if allowed or empty; FALSE otherwise.
   */
  private function isAllowedOrEmpty(EntityInterface $entity, AccountInterface $account) {
    $policy = $this->getCurrentAccessPolicy($entity);
    if ($policy && $this->policySelection->isAllowed($policy, $entity, $account)) {
      return TRUE;
    }

    if ($this->contentAccessPolicyManager->isEmpty($entity)) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Load an Access Policy.
   *
   * @param string $access_policy_id
   *   The Access Policy id.
   *
   * @return \Drupal\access_policy\Entity\AccessPolicyInterface
   *   The Access Policy.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function loadAccessPolicy(string $access_policy_id) {
    /** @var \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy */
    return $this->entityTypeManager->getStorage('access_policy')->load($access_policy_id);
  }

}
