<?php

namespace Drupal\access_policy\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form to configure global Access Policy settings.
 */
class AccessPolicySettingsForm extends ConfigFormBase {

  /**
   * The config.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $config;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a \Drupal\system\ConfigFormBase object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct($config_factory);
    $this->config = $this->config('access_policy.settings');
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'access_policy_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['access_policy.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);

    $form['allow_empty'] = [
      '#title' => $this->t('Allow empty'),
      '#type' => 'checkbox',
      '#default_value' => $this->config->get('allow_empty'),
      '#description' => $this->t('Allow authors to create and edit content without an access policy.'),
    ];

    $form['default_policy'] = [
      '#type' => 'details',
      '#title' => $this->t('Default policy'),
      '#open' => TRUE,
    ];

    $empty_allowed = [':input[name="allow_empty"]' => ['checked' => FALSE]];

    $form['default_policy']['empty'] = [
      '#title' => $this->t('Do not assign a policy'),
      '#parents' => ['default_policy'],
      '#type' => 'radio',
      '#return_value' => '',
      '#default_value' => $this->config->get('default_policy'),
      '#description' => $this->t('New content will not be assigned an access policy.'),
      '#states' => [
        'disabled' => $empty_allowed,
      ],
    ];

    $form['default_policy']['first_available'] = [
      '#title' => $this->t('First available'),
      '#parents' => ['default_policy'],
      '#type' => 'radio',
      '#return_value' => 'first_available',
      '#default_value' => $this->config->get('default_policy'),
      '#description' => $this->t('New content will be assigned the first policy that is available to the user.'),
      '#states' => [
        'disabled' => $empty_allowed,
      ],
    ];
    $form['actions']['#type'] = 'actions';
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save configuration'),
      '#button_type' => 'primary',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config->set('allow_empty', (bool) $form_state->getValue('allow_empty'));
    $this->config->set('default_policy', $form_state->getValue('default_policy'));

    if ($form_state->getValue('allow_empty') == FALSE) {
      $this->config->set('default_policy', 'first_available');
    }
    $this->config->save();
    parent::submitForm($form, $form_state);
  }

}
