<?php

namespace Drupal\access_policy\Plugin\Field;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Field\FieldItemList;
use Drupal\Core\TypedData\ComputedItemListTrait;

/**
 * A computed field that provides a content entity's access policy.
 *
 * It links content entities to an access policy entity via a content access
 * policy content entity.
 */
class AccessPolicyFieldItemList extends FieldItemList {

  use ComputedItemListTrait {
    get as traitGet;
  }

  /**
   * {@inheritdoc}
   */
  protected function computeValue() {
    $access_policy = $this->getAccessPolicyId();
    // Do not store NULL values, in the case where an entity does not have an
    // access policy associated with it, we do not create list items for
    // the computed field.
    if ($access_policy) {
      // An entity can only have a single access policy.
      $this->list[0] = $this->createItem(0, $access_policy);
    }
  }

  /**
   * Gets the access policy ID linked to a content entity.
   *
   * @return string|null
   *   The access policy ID linked to a content entity.
   */
  protected function getAccessPolicyId() {
    $entity = $this->getEntity();

    /** @var \Drupal\access_policy\AccessPolicyInformation $access_policy_info */
    $access_policy_info = \Drupal::service('access_policy.information');
    if (!$access_policy_info->hasEnabledForEntitiesOfEntityType($entity->getEntityType(), $entity->bundle())) {
      return NULL;
    }

    if (!$entity->isNew() && $content_access_policy = $this->loadContentAccessPolicy($entity)) {
      return $content_access_policy->access_policy->value;
    }

  }

  /**
   * Load the content access policy associated with an entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity the access policy entity will be loaded from.
   *
   * @return \Drupal\access_policy\Entity\ContentAccessPolicyInterface|null
   *   The content access policy or FALSE if none exists.
   */
  protected function loadContentAccessPolicy(ContentEntityInterface $entity) {
    $content_access_policy = \Drupal::entityTypeManager()->getStorage('content_access_policy');

    $access_policy = $content_access_policy->getQuery()
      ->accessCheck(FALSE)
      ->condition('content_entity_type_id', $entity->getEntityTypeId())
      ->condition('content_entity_id', $entity->id())
      ->execute();

    if (empty($access_policy)) {
      return NULL;
    }

    /** @var \Drupal\access_policy\Entity\ContentAccessPolicyInterface $content_access_policy */
    $content_access_policy = $content_access_policy->load(key($access_policy));

    return $content_access_policy;
  }

  /**
   * {@inheritdoc}
   */
  public function get($index) {
    if ($index !== 0) {
      throw new \InvalidArgumentException('An entity can not have multiple access policies at the same time.');
    }
    return $this->traitGet($index);
  }

  /**
   * {@inheritdoc}
   */
  public function setValue($values, $notify = TRUE) {
    parent::setValue($values, $notify);
    $this->valueComputed = TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function generateSampleItems($count = 1) {
    // No sample items generated since the starting access policy is always
    // computed based on the default value.
  }

}
