<?php

namespace Drupal\access_policy;

use Drupal\Core\Database\Database;

/**
 * Access policy sql query.
 */
class SqlQuery implements QueryInterface {

  /**
   * Array of joined tables.
   *
   * @var array
   */
  protected $tables = [];

  /**
   * Array of conditions.
   *
   * @var array
   */
  protected $conditions = [];

  /**
   * Flag indicating whether the query is distinct.
   *
   * @var bool
   */
  protected $distinct = FALSE;

  /**
   * The base table.
   *
   * @var string
   */
  protected $baseTable;

  /**
   * The base field.
   *
   * @var string
   */
  protected $baseField;

  /**
   * Initialize the sql query.
   *
   * @param string $base_table
   *   The base table.
   * @param string $base_field
   *   The base field.
   */
  public function init($base_table, $base_field) {
    $this->baseTable = $base_table;
    $this->baseField = $base_field;
  }

  /**
   * {@inheritdoc}
   */
  public function setDistinct() {
    return $this->distinct = TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function isDistinct() {
    return $this->distinct;
  }

  /**
   * {@inheritdoc}
   */
  public function getBaseTable() {
    return $this->baseTable;
  }

  /**
   * {@inheritdoc}
   */
  public function getBaseField() {
    return $this->baseField;
  }

  /**
   * {@inheritdoc}
   */
  public function getConnection() {
    return Database::getConnection();
  }

  /**
   * {@inheritdoc}
   */
  public function getTables() {
    return $this->tables;
  }

  /**
   * {@inheritdoc}
   */
  public function join($table, $alias = NULL, $condition = NULL, $arguments = []) {
    return $this->addJoin('INNER', $table, $alias, $condition, $arguments);
  }

  /**
   * {@inheritdoc}
   */
  public function leftJoin($table, $alias = NULL, $condition = NULL, $arguments = []) {
    return $this->addJoin('LEFT OUTER', $table, $alias, $condition, $arguments);
  }

  /**
   * {@inheritdoc}
   */
  public function addJoin($type, $table, $alias = NULL, $condition = NULL, $arguments = []) {
    if (empty($alias)) {
      $alias = $table;
    }

    $alias_candidate = $alias;
    $count = 2;
    while (!empty($this->tables[$alias_candidate])) {
      $alias_candidate = $alias . '_' . $count++;
    }
    $alias = $alias_candidate;

    if (is_string($condition)) {
      $condition = str_replace('%alias', $alias, $condition);
    }

    $this->tables[$alias] = [
      'join type' => $type,
      'table' => $table,
      'alias' => $alias,
      'condition' => $condition,
      'arguments' => $arguments,
    ];

    return $alias;
  }

  /**
   * {@inheritdoc}
   */
  public function conditions() {
    return $this->conditions;
  }

  /**
   * {@inheritdoc}
   */
  public function condition($field, $value = NULL, $operator = NULL) {
    $this->conditions[] = [
      'field' => $field,
      'value' => $value,
      'operator' => $operator,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function andConditionGroup() {
    return $this->getConnection()->condition('AND');
  }

  /**
   * {@inheritdoc}
   */
  public function orConditionGroup() {
    return $this->getConnection()->condition('OR');
  }

}
