<?php

namespace Drupal\Tests\access_policy\Functional;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\node\Entity\NodeType;

/**
 * Ensures that access policy entity permissions work correctly.
 *
 * @group access_policy
 */
class AccessPolicyEntityAssignTest extends AccessPolicyTestBase {

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'access_policy',
    'access_policy_test',
    'filter',
    'node',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Set to TRUE to strict check all configuration saved.
   *
   * @var bool
   *
   * @see \Drupal\Core\Config\Development\ConfigSchemaChecker
   */
  protected $strictConfigSchema = FALSE;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setup();

    AccessPolicy::create([
      'id' => 'test_policy',
      'label' => 'Test policy',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => 0,
    ])->save();

    AccessPolicy::create([
      'id' => 'test_policy_2',
      'label' => 'Test policy 2',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => -10,
    ])->save();

    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('default_policy', 'first_available')
      ->save();

    \Drupal::service('router.builder')->rebuildIfNeeded();
  }

  /**
   * Tests accessing the access tab for bundle without access policy.
   *
   * In this test a user has the permission to view the Access tab, however the
   * content type doesn't support access policy so that tab shouldn't be
   * accessible.
   */
  public function testAccessEntityWithoutAccessPolicy() {
    // Create an article content type that does not have access policy
    // enabled.
    NodeType::create([
      'type' => 'article',
      'name' => 'Article',
    ])->save();

    $node = $this->drupalCreateNode([
      'type' => 'article',
      'status' => 1,
    ]);

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'edit any article content',
    ]);
    $this->drupalLogin($web_user);

    // They should be able to view the node like normal.
    $this->drupalGet('node/' . $node->id());
    $this->assertSession()->statusCodeEquals(200);

    // They should be able to edit the node like normal.
    $this->drupalGet('node/' . $node->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(403);

    // Make sure that it can't be assigned through the API.
    try {
      $access_policy = \Drupal::entityTypeManager()->getStorage('access_policy')->load('test_policy');
      $this->contentAccessPolicyManager->assign($node, $access_policy);
    }
    catch (\Exception $exception) {
      $message = $exception->getMessage();
      $this->assertEquals('The "test_policy" Access Policy is not enabled for "node" bundle "article".', $message, 'It should have thrown an exception.');
    }

  }

  /**
   * Tests accessing the access tab when it has a forbidden access policy.
   *
   * In this test a user has the permission to view the Access tab, however the
   * content is assigned an access policy that the user doesn't have access
   * to.
   */
  public function testViewAccessTabWithForbiddenAccessPolicy() {

    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    $access_policy = \Drupal::entityTypeManager()->getStorage('access_policy')->load('test_policy_2');
    $this->contentAccessPolicyManager->assign($node, $access_policy);

    $web_user = $this->drupalCreateUser([
      'edit any page content',
      'set entity access policy',
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(403);
  }

  /**
   * Test that the default access policy gets saved to the node when created.
   */
  public function testDefaultAccessPolicyOnNodeCreate() {
    $web_user = $this->drupalCreateUser([
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);
    $node->save();

    $this->assertEquals('test_policy', $node->get('access_policy')->value, 'The access policy should be test_policy');
  }

  /**
   * Tests assigning an access policy to a node from a form.
   */
  public function testAssignAccessPolicyToEntity() {
    // Allow the access policy to be empty.
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('allow_empty', FALSE)
      ->save();

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);
    $this->drupalLogin($web_user);

    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'uid' => $web_user->id(),
    ]);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');

    $edit = [
      'access_policy' => 'test_policy_2',
    ];
    $this->submitForm($edit, 'Save');

    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');

    // Confirm that the content access policy was created when assigned to the
    // node.
    $content_access_policy = $this->entityTypeManager->getStorage('content_access_policy')->loadByProperties([
      'content_entity_type_id' => $node->getEntityTypeId(),
      'content_entity_id' => $node->id(),
      'access_policy' => 'test_policy_2',
    ]);

    $this->assertNotEmpty($content_access_policy, 'The content access policy entity should have been created');

    $node = $this->drupalGetNodeByTitle($node->label());
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node);
    $this->assertEquals('test_policy_2', $policy->id(), 'The access policy was incorrect.');

    // Delete the node and confirm that the content access policy entity was
    // also deleted.
    $node->delete();
    $content_access_policy = $this->entityTypeManager->getStorage('content_access_policy')->loadByProperties([
      'content_entity_type_id' => $node->getEntityTypeId(),
      'content_entity_id' => $node->id(),
    ]);

    $this->assertEmpty($content_access_policy, 'The content access policy entity should have been deleted');

  }

  /**
   * Tests changing an access policy on an entity via a form.
   */
  public function testChangeAccessPolicyOnEntity() {
    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);

    $this->drupalLogin($web_user);

    // Create a published page. Under normal circumstances this will always
    // be visible.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Any time a node is saved it's automatically going to have the default
    // policy stored already. Note that a user has to be logged in to get the
    // default.
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node);
    $this->assertEquals('test_policy_2', $policy->id(), 'The default policy is incorrect.');

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');

    $edit = [
      'access_policy' => 'test_policy',
    ];
    $this->submitForm($edit, 'Save');

    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');

    $node = $this->drupalGetNodeByTitle($node->label());
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node);
    $this->assertEquals('test_policy', $policy->id(), 'The access policy was incorrect.');

  }

  /**
   * Tests removing an access policy from a node via a form.
   */
  public function testSetEmptyAccessPolicyOnEntity() {
    // Allow the access policy to be empty.
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('allow_empty', TRUE)
      ->save();

    // Create a published page. Under normal circumstances this will always
    // be visible.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Assign test_policy access policy to the node.
    $access_policy = \Drupal::entityTypeManager()->getStorage('access_policy')->load('test_policy');
    $this->contentAccessPolicyManager->assign($node, $access_policy);

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    // Assert that the "- Select -" option exists in the select list.
    $this->assertSession()->pageTextContains('- Select -');

    // Remove the access policy and then assert that the access policy is
    // marked as removed and it won't revert to the default value again.
    $edit = [
      'access_policy' => '',
    ];
    $this->submitForm($edit, 'Save');

    $node = $this->drupalGetNodeByTitle($node->label());
    $empty = $this->contentAccessPolicyManager->isEmpty($node);

    $this->assertTrue($empty, 'The access policy should have been removed.');
    $this->assertSession()->fieldValueEquals('access_policy', '');
  }

  /**
   * Tests not allowing users to disable an access policy on a node.
   */
  public function testDoNotRemoveAccessPolicyOnEntityWhenNotAllowed() {
    // Allow the access policy to be disabled.
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('allow_empty', FALSE)
      ->save();

    // Create a published page. Under normal circumstances this will always
    // be visible.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Assign test_policy access policy to the node.
    $access_policy = \Drupal::entityTypeManager()->getStorage('access_policy')->load('test_policy');
    $this->contentAccessPolicyManager->assign($node, $access_policy);

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    // Assert that the "Disabled" option exists in the select list.
    $this->assertSession()->pageTextNotContains('- Select -');

    // Attempt to remove the access policy. This should not be allowed.
    try {
      $edit = [
        'access_policy' => '',
      ];
      $this->submitForm($edit, 'Save');
    }
    catch (\InvalidArgumentException $e) {
      // Expected exception; just continue testing.
    }

    $node = $this->drupalGetNodeByTitle($node->label());
    $empty = $this->contentAccessPolicyManager->isEmpty($node);

    $this->assertFalse($empty, 'The access policy should not have been removed.');
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');
  }

}
