<?php

namespace Drupal\Tests\access_policy\Functional;

use Drupal\Tests\field\Traits\EntityReferenceTestTrait;
use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\Tests\user\Traits\UserCreationTrait;

/**
 * Ensures that listings only display entities that a user is allowed to view.
 *
 * @group access_policy
 */
class AccessRuleEntityFieldTest extends AccessPolicyTestBase {

  use UserCreationTrait;
  use EntityReferenceTestTrait;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'access_policy',
    'access_policy_test',
    'filter',
    'node',
    'field',
    'user',
    'datetime',
    'views',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Set to TRUE to strict check all configuration saved.
   *
   * @var bool
   *
   * @see \Drupal\Core\Config\Development\ConfigSchemaChecker
   */
  protected $strictConfigSchema = FALSE;

  /**
   * Tests viewing a node that has hardcoded values in the access rule.
   */
  public function testEntityFieldStandardWithValue() {
    $policy = AccessPolicy::create([
      'id' => 'matching_string',
      'label' => 'Matching String Access Policy',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);
    $policy->save();
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Foo',
      'operator' => '=',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $node_1 = $this->drupalCreateNode([
      'title' => 'Node 1 (With text: Foo)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Foo',
      ],
    ]);

    $node_2 = $this->drupalCreateNode([
      'title' => 'Node 2 (With text: Bar)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Bar',
      ],
    ]);

    $node_3 = $this->drupalCreateNode([
      'title' => 'Node 3 (Without text)',
      'type' => 'page',
      'status' => 1,
    ]);

    $this->assignAccessPolicy($node_1, 'matching_string');
    $this->assignAccessPolicy($node_2, 'matching_string');
    $this->assignAccessPolicy($node_3, 'matching_string');

    // We're not comparing field values with a user so it's not necessary to set
    // it.
    $admin_user = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view matching_string content',
    ]);
    $admin_user->save();

    $this->drupalLogin($admin_user);
    $this->drupalGet('node/' . $node_1->id());
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('node/' . $node_2->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('node/' . $node_3->id());
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce('Node 1 (With text: Foo)');
    $this->assertSession()->pageTextNotContains('Node 2 (With text: Bar)');
    $this->assertSession()->pageTextContainsOnce('Node 3 (Without text)');
  }

  /**
   * Tests the field value access rule with entity reference fields.
   */
  public function testEntityFieldEntityReferenceWithValue() {
    $this->createEntityReferenceField('node', 'page', 'field_entity_reference', 'Entity Reference', 'node', 'default', [], -1);

    $policy = AccessPolicy::create([
      'id' => 'match_reference_value',
      'label' => 'Matching entity reference with value',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_entity_reference');
    $access_rule->setSettings([
      'value' => [
        ['target_id' => 1],
      ],
      'operator' => 'in',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $referenced_node_1 = $this->drupalCreateNode([
      'nid' => 1,
      'title' => 'Referenced Node #1',
      'type' => 'page',
      'status' => 1,
    ]);

    $referenced_node_2 = $this->drupalCreateNode([
      'title' => 'Referenced Node #1',
      'type' => 'page',
      'status' => 1,
    ]);

    // This is referencing node 1 in the access policy and should be visible.
    $node_1 = $this->drupalCreateNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
      'field_entity_reference' => [
        $referenced_node_1,
      ],
    ]);

    // This is referencing a different node and shoud not be visibe.
    $node_2 = $this->drupalCreateNode([
      'title' => 'Node 2',
      'type' => 'page',
      'status' => 1,
      'field_entity_reference' => [
        $referenced_node_2,
      ],
    ]);

    // This has no value and will therefore be visible.
    $node_3 = $this->drupalCreateNode([
      'title' => 'Node 3',
      'type' => 'page',
      'status' => 1,
    ]);

    $this->assignAccessPolicy($node_1, 'match_reference_value');
    $this->assignAccessPolicy($node_2, 'match_reference_value');
    $this->assignAccessPolicy($node_3, 'match_reference_value');

    $web_user = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view match_reference_value content',
    ]);

    // We're not matching values on the user, so we don't need to assign it.
    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node_1->id());
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('node/' . $node_2->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('node/' . $node_3->id());
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce('Node 1');
    $this->assertSession()->pageTextNotContains('Node 2');
    $this->assertSession()->pageTextContains('Node 3');

  }

}
