<?php

namespace Drupal\Tests\access_policy\Functional;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\Core\Entity\EntityInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;

/**
 * Ensures that access policy entity permissions work correctly.
 *
 * @group access_policy
 */
class AccessRuleWidgetTest extends AccessPolicyTestBase {

  use TaxonomyTestTrait;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'access_policy',
    'access_policy_test',
    'taxonomy',
    'filter',
    'node',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Set to TRUE to strict check all configuration saved.
   *
   * @var bool
   *
   * @see \Drupal\Core\Config\Development\ConfigSchemaChecker
   */
  protected $strictConfigSchema = FALSE;

  /**
   * The vocabulary.
   *
   * @var \Drupal\taxonomy\VocabularyInterface
   */
  protected $vocabulary;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setup();

    // Then create some taxonomy terms that are assigned their own policy.
    $this->vocabulary = Vocabulary::create([
      'name' => 'department',
      'vid' => 'department',
    ]);
    $this->vocabulary->save();

    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_term_reference',
      'entity_type' => 'node',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_term_reference',
      'bundle' => 'page',
      'translatable' => FALSE,
    ]);
    $field->save();

    AccessPolicy::create([
      'id' => 'test_policy',
      'label' => 'Test policy',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => 0,
    ])->save();

    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('default_policy', 'first_available')
      ->save();

    \Drupal::service('router.builder')->rebuildIfNeeded();
  }

  /**
   * Tests that an access rule with a form renders correctly on the access tab.
   */
  public function testAccessRuleWithTextFieldWidget() {

    $policy = AccessPolicy::create([
      'id' => 'test_policy_with_form',
      'label' => 'Test policy with form',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => -10,
    ]);

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Some text',
      'operator' => '=',
      'widget' => [
        'show' => TRUE,
      ],
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Create a published page. Under normal circumstances this will always
    // be visible.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Assign test_policy access policy to the node.
    $access_policy = \Drupal::entityTypeManager()->getStorage('access_policy')->load('test_policy_with_form');
    $this->contentAccessPolicyManager->assign($node, $access_policy);

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy_with_form access policy',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldExists('field_text[0][value]');

    $edit = [
      'field_text[0][value]' => 'My matching text',
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->fieldValueEquals('field_text[0][value]', 'My matching text');

    // Reload the node and all its values.
    $node = $this->drupalGetNodeByTitle($node->label());
    $this->assertEquals('My matching text', $node->get('field_text')->getString(), 'Value not saved on the entity');
  }

  /**
   * Tests an access rule with entity reference field widget.
   */
  public function testAccessRuleWithEntityReferenceWidget() {

    $policy = AccessPolicy::create([
      'id' => 'test_policy_with_form',
      'label' => 'Test policy with form',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => -10,
    ]);

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_term_reference');
    $access_rule->setSettings([
      'value' => 'Some text',
      'operator' => '=',
      'widget' => [
        'show' => TRUE,
      ],
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Create a published page. Under normal circumstances this will always
    // be visible.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Assign test_policy access policy to the node.
    $this->contentAccessPolicyManager->assign($node, $policy);

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy_with_form access policy',
    ]);
    $this->drupalLogin($web_user);

    $term_1 = $this->createTerm($this->vocabulary, [
      'name' => 'Term 1',
      'status' => 1,
    ]);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    $value = $term_1->label() . ' (' . $term_1->id() . ')';
    $edit = [
      'field_term_reference[0][target_id]' => $value,
    ];
    $this->submitForm($edit, 'Save');

    // Reload the form and confirm that it was saved successfully.
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->fieldValueEquals('field_term_reference[0][target_id]', $value);
    $this->assertEquals([$term_1->id()], $this->getNodeTerms($node), 'The taxonomy terms were not saved correctly');

    // Change the widget to entity reference auto complete. and confirm that it
    // still renders and saves successfully.
    $rule = $policy->getAccessRule('field_term_reference');
    $settings = $rule['settings'];
    $settings['widget']['settings'] = [
      'field_widget' => 'entity_reference_autocomplete_tags',
    ];
    $policy->updateAccessRule('field_term_reference', $settings);
    $policy->save();

    $term_2 = $this->createTerm($this->vocabulary, [
      'name' => 'Term 2',
      'status' => 1,
    ]);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    $value = $term_2->label() . ' (' . $term_2->id() . ')';
    $edit = [
      'field_term_reference[target_id]' => $value,
    ];
    $this->submitForm($edit, 'Save');

    // Reload the form and confirm that it was saved successfully.
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->fieldValueEquals('field_term_reference[target_id]', $value);
    $this->assertEquals([$term_1->id()], $this->getNodeTerms($node), 'The taxonomy terms were not saved correctly');
  }

  /**
   * Load the terms on the node.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return array
   *   Array of term ids.
   */
  protected function getNodeTerms(EntityInterface $entity) {
    $node = $this->drupalGetNodeByTitle($entity->label());
    $terms = $node->get('field_term_reference')->referencedEntities();
    $items = [];
    foreach ($terms as $term) {
      $items[] = $term->id();
    }
    return $items;
  }

}
