<?php

namespace Drupal\Tests\access_policy\Kernel;

use Drupal\node\Entity\NodeType;
use Drupal\access_policy\Entity\AccessPolicy;

/**
 * Tests Access Policy selection manaager.
 *
 * @group access_policy
 */
class AccessPolicySelectionTest extends AccessPolicyKernelTestBase {

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  protected function setUp(): void {
    parent::setUp();

    NodeType::create(['type' => 'page', 'name' => 'Page']);
    NodeType::create(['type' => 'article', 'name' => 'Article']);
  }

  /**
   * Tests getting the default access policy for first available policies.
   *
   * If the default setting is set to first_available it should return the
   * first access policy that the user has access too. Prioritized by
   * weight.
   *
   * @throws \Exception
   */
  public function testGetFirstAccessibleDefaultAccessPolicy() {
    // Set access policies as first accessible.
    $config = $this->config('access_policy.settings');
    $config->set('default_policy', 'first_available');
    $config->save();

    $accessPolicySelection = \Drupal::service('access_policy.selection');

    AccessPolicy::create([
      'id' => 'policy_test_1',
      'label' => 'Policy test 1',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => 0,
    ])->save();

    AccessPolicy::create([
      'id' => 'policy_test_2',
      'label' => 'Policy test 2',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => -10,
    ])->save();

    $node = $this->createNode([
      'type' => 'page',
      'title' => 'Page',
    ]);
    $node->save();

    // This user doesn't have permission to any of these access policies so it
    // should not return anything.
    $user = $this->createUser();
    $default_policy = $accessPolicySelection->getDefaultPolicy($node, $user);
    $this->assertEquals(FALSE, $default_policy, 'This should not return an access policy');

    // The user only has access to editing one access policy therefore it
    // should only return that access policy.
    $user_2 = $this->createUser([
      'assign policy_test_2 access policy',
    ]);
    $default_policy = $accessPolicySelection->getDefaultPolicy($node, $user_2);
    $this->assertEquals('policy_test_2', $default_policy->id(), 'Only one access policy is assigned.');

    // The user has access to two access policies but it will only return the
    // one with the lowest weight (highest priority).
    $user_3 = $this->createUser([
      'assign policy_test_1 access policy',
      'assign policy_test_2 access policy',
    ]);
    $default_policy = $accessPolicySelection->getDefaultPolicy($node, $user_3);
    $this->assertEquals('policy_test_2', $default_policy->id(), 'policy test 2 has lower weight (higher priority) and should be returned.');
  }

  /**
   * Tests getting the first available policy across bundles.
   *
   * If a user has access to two policies, but they're enabled on different
   * bundles then make sure it returns the right default policy for each bundle.
   */
  public function testGetFirstAvailableAccessPolicyWithMultipleBundles() {
    // Set access policies as first accessible.
    $config = $this->config('access_policy.settings');
    $config->set('default_policy', 'first_available');
    $config->save();

    $accessPolicySelection = \Drupal::service('access_policy.selection');

    AccessPolicy::create([
      'id' => 'policy_test_1',
      'label' => 'Policy test 1',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => 0,
    ])->save();

    AccessPolicy::create([
      'id' => 'policy_test_2',
      'label' => 'Policy test 2',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['article'],
      'weight' => -10,
    ])->save();

    $user = $this->createUser([
      'assign policy_test_1 access policy',
      'assign policy_test_2 access policy',
    ]);

    // This should return policy_test_1 since it's the only one enabled for that
    // policy.
    $node = $this->createNode([
      'type' => 'page',
      'title' => 'Page',
    ]);
    $node->save();

    $default_policy = $accessPolicySelection->getDefaultPolicy($node, $user);
    $this->assertEquals('policy_test_1', $default_policy->id(), 'It should have returned policy_test_1.');

    // This should return policy_test_2 since it's the only one enabled for that
    // policy.
    $node_2 = $this->createNode([
      'type' => 'article',
      'title' => 'Page',
    ]);
    $node->save();

    $default_policy = $accessPolicySelection->getDefaultPolicy($node_2, $user);
    $this->assertEquals('policy_test_2', $default_policy->id(), 'It should have returned policy_test_2');
  }

  /**
   * Tests getting the default access policy when default is empty.
   *
   * If the default setting is set to empty the user should not get a default
   * policy. Even if they have permission to use it.
   *
   * @throws \Exception
   */
  public function testGetDefaultAccessPolicyWhenEmpty() {
    $config = $this->config('access_policy.settings');
    $config->set('default_policy', '');
    $config->save();

    $accessPolicySelection = \Drupal::service('access_policy.selection');

    AccessPolicy::create([
      'id' => 'policy_test',
      'label' => 'Policy test',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
      'weight' => 0,
    ])->save();

    $node = $this->createNode([
      'type' => 'page',
      'title' => 'Page',
    ]);
    $node->save();

    $user = $this->createUser([
      'assign policy_test access policy',
    ]);
    $default_policy = $accessPolicySelection->getDefaultPolicy($node, $user);
    $this->assertEquals(FALSE, $default_policy, 'This should not return an access policy');
  }

}
