<?php

namespace Drupal\Tests\access_policy\Kernel\Plugin\AccessRule;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\Plugin\AccessRule\AccessRuleOperatorTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\access_policy\Kernel\AccessPolicyKernelTestBase;

/**
 * Tests EntityField base plugin features.
 *
 * @group access_policy
 */
class EntityFieldBaseTest extends AccessPolicyKernelTestBase {

  use TaxonomyTestTrait;
  use AccessRuleOperatorTrait;

  /**
   * The tags vocabulary.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  protected $vocabulary;

  /**
   * The modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'taxonomy',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy_test']);
    $this->installEntitySchema('taxonomy_term');

    // Create a tags vocabulary.
    $this->vocabulary = Vocabulary::create([
      'name' => 'tags',
      'vid' => 'tags',
    ]);
    $this->vocabulary->save();

    // Create the taxonomy term entity reference field for the node and user.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_tags',
      'entity_type' => 'user',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_tags',
      'bundle' => 'user',
      'translatable' => FALSE,
    ]);
    $field->save();

    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_tags',
      'entity_type' => 'node',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_tags',
      'bundle' => 'page',
      'translatable' => FALSE,
    ]);
    $field->save();

    // Creating a policy is necessary to register the plugins.
    $policy = AccessPolicy::create([
      'id' => 'basic_policy',
      'label' => 'Basic test policy',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);
    $policy->save();
  }

  /**
   * Tests access for entities with empty values.
   */
  public function testEntityFieldBaseWithEmptyValue() {
    $policy = AccessPolicy::create([
      'id' => 'policy_empty_fields',
      'label' => 'Policy with empty fields',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);
    $policy->save();

    // Access rule from a hard coded value.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Foo',
      'operator' => 'contains',
      'query' => TRUE,
      'empty_behavior' => 'deny',
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Node 1 does not have any field values.
    $node_1 = $this->createNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
    ]);

    $this->assignAccessPolicy($node_1, 'policy_empty_fields');

    $web_user = $this->createUser([
      'access content',
      'view policy_empty_fields content',
    ]);
    $web_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $web_user, 'view');
    $this->assertFalse($access, 'This user should not have access');

    $this->setCurrentUser($web_user);
    $this->assertQueryResults('node', []);

    // Change the access rule settings to allow acccess.
    $access_rule_settings = $access_rule->getSettings();
    $access_rule_settings['empty_behavior'] = 'allow';
    $policy->updateAccessRule('field_text', $access_rule_settings);
    $policy->save();

    $access = $this->accessPolicyValidator->validate($node_1, $web_user, 'view');
    $this->assertTrue($access, 'This user should have access');

    $this->setCurrentUser($web_user);
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);
  }

  /**
   * Tests access rule validation for overridden operations.
   */
  public function testEntityFieldBaseWithSpecificOperations() {
    $policy = AccessPolicy::create([
      'id' => 'policy_op_override',
      'label' => 'Policy with overridden operations',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);
    $policy->save();

    // Access rule that validates during the "update" operation. This means that
    // All other operations should pass except for edit.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Foo',
      'operator' => '=',
      'admin_label' => '',
      'operations' => ['update'],
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Node 1 does not have any field values.
    $node_1 = $this->createNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Bar',
    ]);

    $this->assignAccessPolicy($node_1, 'policy_op_override');

    $web_user = $this->createUser([
      'access content',
      'view policy_op_override content',
    ]);
    $web_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $web_user, 'view');
    $this->assertTrue($access, 'This user should have access');

    $access = $this->accessPolicyValidator->validate($node_1, $web_user, 'update');
    $this->assertFalse($access, 'This user should not have access');
  }

  /**
   * Basic test for testing access_rule_data_alter.
   */
  public function testAccessRuleDataAlter() {
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');

    // Check the access rule description, confirm that hook_access_rule_alter()
    // works.
    // @see access_policy_test.module.
    $description = $access_rule->getDefinition()->getDescription();
    $this->assertEquals('Foo bar', $description);
  }

}
