<?php

namespace Drupal\Tests\access_policy\Kernel\Plugin\AccessRule;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\Plugin\AccessRule\AccessRuleOperatorTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\Tests\node\Traits\NodeCreationTrait;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\access_policy\Kernel\AccessPolicyKernelTestBase;

/**
 * Tests FieldMatchEntityReference plugin features.
 *
 * @group access_policy
 */
class FieldMatchEntityReferenceTest extends AccessPolicyKernelTestBase {

  use TaxonomyTestTrait;
  use NodeCreationTrait;
  use AccessRuleOperatorTrait;

  /**
   * The tags vocabulary.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  protected $vocabulary;

  /**
   * The modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'taxonomy',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy_test']);
    $this->installEntitySchema('taxonomy_term');

    // Create a tags vocabulary.
    $this->vocabulary = Vocabulary::create([
      'name' => 'tags',
      'vid' => 'tags',
    ]);
    $this->vocabulary->save();

    // Create the taxonomy term entity reference field for the node and user.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_tags',
      'entity_type' => 'user',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_tags',
      'bundle' => 'user',
      'translatable' => FALSE,
    ]);
    $field->save();

    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_tags',
      'entity_type' => 'node',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_tags',
      'bundle' => 'page',
      'translatable' => FALSE,
    ]);
    $field->save();

    // Creating a policy is necessary to register the plugins.
    $policy = AccessPolicy::create([
      'id' => 'basic_policy',
      'label' => 'Basic test policy',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);
    $policy->save();

  }

  /**
   * Tests the field value access rule with not in operator.
   */
  public function testFieldMatchEntityReference() {
    // Test entity reference fields.
    $term_1 = $this->createTerm($this->vocabulary, [
      'title' => 'Term 1',
    ]);
    $term_1->save();

    $term_2 = $this->createTerm($this->vocabulary, [
      'name' => 'Term 2',
    ]);
    $term_2->save();

    $policy = AccessPolicy::create([
      'id' => 'match_reference_value',
      'label' => 'Matching entity reference between node and user',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('node', 'match_field_tags');
    $access_rule->setSettings([
      'operator' => 'in',
      'value' => ['field' => 'field_tags'],
      'query' => TRUE,
    ]);

    $policy->addAccessRule($access_rule);
    $policy->save();

    // This is referencing term 1 and should  visible.
    $node_1 = $this->createNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
      'field_tags' => [
        $term_1,
      ],
    ]);

    // This is referencing term 2 and should not be visible.
    $node_2 = $this->createNode([
      'title' => 'Node 1',
      'type' => 'page',
      'status' => 1,
      'field_tags' => [
        $term_2,
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'match_reference_value');
    $this->assignAccessPolicy($node_2, 'match_reference_value');

    $web_user = $this->createUser([
      'access content',
      'view match_reference_value content',
    ], NULL, FALSE, ['field_tags' => $term_1]);

    $this->setCurrentUser($web_user);
    $access = $this->accessPolicyValidator->validate($node_1, $web_user, 'view');
    $this->assertTrue($access, 'The user should have access');

    $access = $this->accessPolicyValidator->validate($node_2, $web_user, 'view');
    $this->assertFalse($access, 'The user should not have access');
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);
  }

  /**
   * Tests that a user has access to content if they have a reference to the
   * original author.
   */
  public function testCurrentUserHasReferenceToOriginalAuthor() {
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_teacher',
      'entity_type' => 'user',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'user',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_teacher',
      'bundle' => 'user',
      'translatable' => FALSE,
    ]);
    $field->save();

    $policy = $this->createAccessPolicy('author_reference', [
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);

    $access_rule = $this->accessRuleManager->getHandler('node', 'author_reference');
    $access_rule->setSettings([
      'value' => ['field' => 'field_teacher'],
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $teacher = $this->createUser();

    $node_1 = $this->createNode([
      'title' => 'Node authored by teacher',
      'uid' => $teacher->id(),
    ]);
    $this->assignAccessPolicy($node_1, 'author_reference');

    $student = $this->createUser([
      'view author_reference content',
    ]);
    $student->set('field_teacher', [
      $teacher,
    ]);
    $student->save();

    // @todo it's confusing that you have to set the current user.
    $this->setCurrentUser($student);
    $access = $this->accessPolicyValidator->validate($node_1, $student, 'view');
    $this->assertTrue($access, 'The user should have access');
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);

    // This student does not have access to the content and they should not have
    // access.
    $student_2 = $this->createUser([
      'view author_reference content',
    ]);
    $this->setCurrentUser($student_2);
    $access = $this->accessPolicyValidator->validate($node_1, $student_2, 'view');
    $this->assertFalse($access, 'The user should not have access');
    $this->assertQueryResults('node', []);
  }

  /**
   * Tests that a user has access to a content if referencing it directly.
   */
  public function testCurrentUserHasReferenceToContentItem() {
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_node',
      'entity_type' => 'user',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'node',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_node',
      'bundle' => 'user',
      'translatable' => FALSE,
    ]);
    $field->save();

    $policy = $this->createAccessPolicy('node_reference', [
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);

    $access_rule = $this->accessRuleManager->getHandler('node', 'entity_reference');
    $access_rule->setSettings([
      'value' => ['field' => 'field_node'],
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $node_1 = $this->createNode([
      'title' => 'Node 1',
    ]);
    $this->assignAccessPolicy($node_1, 'node_reference');

    $web_user = $this->createUser([
      'view node_reference content',
    ]);
    $web_user->set('field_node', [
      $node_1,
    ]);
    $web_user->save();

    // @todo it's confusing that you have to set the current user.
    $this->setCurrentUser($web_user);
    $access = $this->accessPolicyValidator->validate($node_1, $web_user, 'view');
    $this->assertTrue($access, 'The user should have access');
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);

    // This student does not have acess to the content and they should not have
    // access.
    $web_user_2 = $this->createUser([
      'view node_reference content',
    ]);
    $this->setCurrentUser($web_user_2);
    $access = $this->accessPolicyValidator->validate($node_1, $web_user_2, 'view');
    $this->assertFalse($access, 'The user should not have access');
    $this->assertQueryResults('node', []);
  }

  /**
   * Tests that a user has access to a content if referencing it directly.
   */
  public function testContentItemHasReferenceToCurrentUser() {
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_user',
      'entity_type' => 'node',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'user',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_user',
      'bundle' => 'page',
      'translatable' => FALSE,
    ]);
    $field->save();

    $policy = $this->createAccessPolicy('user_reference', [
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ]);

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_user_user_reference');
    $policy->addAccessRule($access_rule);
    $policy->save();

    $web_user = $this->createUser([
      'view user_reference content',
    ]);
    $node_1 = $this->createNode([
      'title' => 'Node 1',
      'field_user' => [
        $web_user,
      ],
    ]);
    $this->assignAccessPolicy($node_1, 'user_reference');

    // @todo it's confusing that you have to set the current user.
    $this->setCurrentUser($web_user);
    $access = $this->accessPolicyValidator->validate($node_1, $web_user, 'view');
    $this->assertTrue($access, 'The user should have access');
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);

    $web_user_2 = $this->createUser([
      'view user_reference content',
    ]);
    $this->setCurrentUser($web_user_2);
    $access = $this->accessPolicyValidator->validate($node_1, $web_user_2, 'view');
    $this->assertFalse($access, 'The user should not have access');
    $this->assertQueryResults('node', []);
  }

}
